from __future__ import annotations

import logging
from pathlib import Path
from typing import List, Generator, Dict, Set, cast

from genomics_data_index.storage.io.FeaturesReader import FeaturesReader
from genomics_data_index.storage.io.SampleData import SampleData
from genomics_data_index.storage.io.SampleDataPackage import SampleDataPackage
from genomics_data_index.storage.io.SampleFilesProcessor import SampleFilesProcessor
from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.NucleotideSampleDataBedMask import NucleotideSampleDataBedMask
from genomics_data_index.storage.io.mutation.NucleotideSampleDataSequenceMask import NucleotideSampleDataSequenceMask
from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile
from genomics_data_index.storage.io.mutation.VcfVariantsReader import VcfVariantsReader
from genomics_data_index.storage.io.mutation.variants_processor.SerialVcfVariantsTableProcessor import \
    SerialVcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.mutation.variants_processor.VcfVariantsTableProcessor import \
    VcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.processor.NullSampleFilesProcessor import NullSampleFilesProcessor

logger = logging.getLogger(__file__)


class NucleotideSampleDataPackage(SampleDataPackage):

    def __init__(self, sample_data_dict: Dict[str, NucleotideSampleData],
                 sample_files_processor: SampleFilesProcessor,
                 variants_processor_factory: VcfVariantsTableProcessorFactory,
                 index_unknown_missing: bool = True):
        super().__init__(index_unknown_missing=index_unknown_missing)
        self._sample_data_dict = sample_data_dict
        self._sample_files_processor = sample_files_processor
        self._variants_processor_factory = variants_processor_factory
        self._sample_names = set(self._sample_data_dict.keys())

    def get_features_reader(self) -> FeaturesReader:
        return VcfVariantsReader.create(sample_files_map=self._sample_data_dict,
                                        variants_processor_factory=self._variants_processor_factory,
                                        include_masked_regions=self.index_unknown_missing())

    def sample_names(self) -> Set[str]:
        return self._sample_names

    def iter_sample_data(self) -> Generator[SampleData, None, None]:
        return self._sample_files_processor.process(list(self._sample_data_dict.values()))

    def get_sample_data(self) -> Dict[str, NucleotideSampleData]:
        return self._sample_data_dict

    def get_variants_processor_factory(self) -> VcfVariantsTableProcessorFactory:
        return self._variants_processor_factory

    def get_sample_files_processor(self) -> SampleFilesProcessor:
        return self._sample_files_processor

    def process_all_data(self) -> SampleDataPackage:
        processed_data = {}
        for sample_data in self.iter_sample_data():
            processed_data[sample_data.sample_name] = sample_data

        processed_data = cast(Dict[str, NucleotideSampleData], processed_data)
        return NucleotideSampleDataPackage.create_from_sample_data(sample_data_dict=processed_data,
                                                                   sample_files_processor=self._sample_files_processor,
                                                                   variants_processor_factory=self._variants_processor_factory,
                                                                   index_unknown_missing=self.index_unknown_missing())

    @classmethod
    def mask_is_bed(cls, file: Path) -> bool:
        return file.name.endswith('.bed') or file.name.endswith('.bed.gz')

    @classmethod
    def mask_is_sequence(cls, file: Path) -> bool:
        return SequenceFile(file).is_fasta()

    @classmethod
    def create_from_vcf_masks(cls, sample_vcf_map: Dict[str, Path],
                              masked_genomic_files_map: Dict[str, Path] = None,
                              sample_files_processor: SampleFilesProcessor = NullSampleFilesProcessor.instance(),
                              variants_processor_factory: VcfVariantsTableProcessorFactory = SerialVcfVariantsTableProcessorFactory.instance(),
                              index_unknown_missing: bool = True) -> NucleotideSampleDataPackage:
        if masked_genomic_files_map is None:
            masked_genomic_files_map = {}

        sample_data_dict = {}
        for sample_name in sample_vcf_map:
            vcf_file = sample_vcf_map[sample_name]
            if sample_name in masked_genomic_files_map:
                mask_file = masked_genomic_files_map[sample_name]
            else:
                mask_file = None

            if mask_file is None or cls.mask_is_bed(mask_file):
                sample_data = NucleotideSampleDataBedMask.create(
                    sample_name=sample_name,
                    vcf_file=vcf_file,
                    mask_bed_file=mask_file
                )
            elif cls.mask_is_sequence(mask_file):
                sample_data = NucleotideSampleDataSequenceMask.create(
                    sample_name=sample_name,
                    vcf_file=vcf_file,
                    sample_mask_sequence=mask_file
                )
            else:
                raise Exception(f'Unsupported mask file=[{mask_file}]. Must be either BED or FASTA.')

            sample_data_dict[sample_name] = sample_data

        return NucleotideSampleDataPackage(sample_data_dict=sample_data_dict,
                                           sample_files_processor=sample_files_processor,
                                           variants_processor_factory=variants_processor_factory,
                                           index_unknown_missing=index_unknown_missing
                                           )

    @classmethod
    def create_from_snippy(cls, sample_dirs: List[Path],
                           sample_files_processor: SampleFilesProcessor = NullSampleFilesProcessor.instance(),
                           variants_processor_factory: VcfVariantsTableProcessorFactory = SerialVcfVariantsTableProcessorFactory.instance(),
                           index_unknown_missing: bool = True) -> NucleotideSampleDataPackage:

        sample_data_dict = {}
        subtract_vcf_from_mask = True
        logger.debug(f'Creating SampleDataPackage for snippy data. Setting '
                     f'subtract_vcf_from_mask={subtract_vcf_from_mask} to account for the deletions that are '
                     f'in the VCF file and also in the sequence alignment (snps.aligned.fa) file.')
        for d in sample_dirs:
            sample_name = d.name
            sample_data = NucleotideSampleDataSequenceMask.create(
                sample_name=sample_name,
                vcf_file=Path(d, 'snps.vcf.gz'),
                sample_mask_sequence=Path(d, 'snps.aligned.fa'),
                subtract_vcf_from_mask=subtract_vcf_from_mask,
            )
            sample_data_dict[sample_name] = sample_data

        return NucleotideSampleDataPackage(sample_data_dict=sample_data_dict,
                                           sample_files_processor=sample_files_processor,
                                           variants_processor_factory=variants_processor_factory,
                                           index_unknown_missing=index_unknown_missing
                                           )

    @classmethod
    def create_from_sample_data(cls, sample_data_dict: Dict[str, NucleotideSampleData],
                                sample_files_processor: SampleFilesProcessor = NullSampleFilesProcessor.instance(),
                                variants_processor_factory: VcfVariantsTableProcessorFactory = SerialVcfVariantsTableProcessorFactory.instance(),
                                index_unknown_missing: bool = True) -> NucleotideSampleDataPackage:
        return NucleotideSampleDataPackage(sample_data_dict=sample_data_dict,
                                           sample_files_processor=sample_files_processor,
                                           variants_processor_factory=variants_processor_factory,
                                           index_unknown_missing=index_unknown_missing)



import tempfile
from pathlib import Path

from sourmash import load_file_as_signatures

from genomics_data_index.storage.index.KmerIndexer import KmerIndexManager
from genomics_data_index.storage.index.KmerIndexer import KmerIndexerSourmash
from genomics_data_index.test.integration import data_dir
from genomics_data_index.test.integration.storage.index import get_values_from_signatures


def test_index_multiple_files():
    sampleA = data_dir / 'SampleA' / 'snps.aligned.nogap.fa.gz'
    sampleB = data_dir / 'SampleB' / 'snps.aligned.nogap.fa.gz'

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        genomes_files = [
            ('SampleA', [sampleA]),
            ('SampleB', [sampleB])
        ]

        kmer_indexer = KmerIndexerSourmash(k=31, scaled=1000, compress=True)
        kmer_index_manager = KmerIndexManager(tmp_path, kmer_indexer=kmer_indexer)
        indexed_files = kmer_index_manager.index_all_genomes(genomes_files)

        assert {'SampleA', 'SampleB'} == set(indexed_files.keys())

        sigsA = load_file_as_signatures(str(indexed_files['SampleA']))
        valuesA = get_values_from_signatures(sigsA)
        assert {31} == valuesA['ksize']
        assert {1000} == valuesA['scaled']

        sigsB = load_file_as_signatures(str(indexed_files['SampleB']))
        valuesB = get_values_from_signatures(sigsB)
        assert {31} == valuesB['ksize']
        assert {1000} == valuesB['scaled']



#!/usr/bin/env python
# Prints if the given position (1-based, like in VCF files) is in the given bed files

import sys
from pathlib import Path

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions

if len(sys.argv) < 2:
    print(f"Usage: {sys.argv[0]} [position] [file.bed] ...\n")
    sys.exit(1)

seq_position = sys.argv[1]
if ':' in seq_position:
    sequence, position = seq_position.split(':')
    position = int(position)
else:
    sequence = 'reference'
    position = int(seq_position)

for file in sys.argv[2:]:
    mask = MaskedGenomicRegions.from_file(Path(file))
    if mask.contains(sequence=sequence, position=position, start_position_index='1'):
        print(f'{file} contains {sequence}:{position}')
#    else:
#        print(f'{file} does not contain {sequence}:{position}')



import re
import tempfile
from pathlib import Path
from typing import Dict

from Bio import SeqIO

from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile
from genomics_data_index.test.integration import reference_file_5000_snpeff, reference_file_5000_snpeff_2, \
    reference_file, reference_file_lzma, reference_file_bzip2


def parse_snpeff_config(config_file: Path) -> Dict[str, str]:
    key_values = {}
    with open(config_file, 'r') as fh:
        for line in fh:
            line = line.strip()
            if not line.startswith('#'):
                tokens = re.split(r'[:=]', line)

                if len(tokens) == 2:
                    key_values[tokens[0].strip()] = tokens[1].strip()

    return key_values


def test_read_fasta_file():
    sequence_file = SequenceFile(reference_file)
    name, records = sequence_file.parse_sequence_file()

    assert 'genome' == name
    assert 1 == len(records)
    record = records[0]
    assert 'reference' == record.id
    assert 5180 == len(record)


def test_read_fasta_file_lzma():
    sequence_file = SequenceFile(reference_file_lzma)
    name, records = sequence_file.parse_sequence_file()

    assert 'genome' == name
    assert 1 == len(records)
    record = records[0]
    assert 'reference' == record.id
    assert 5180 == len(record)


def test_read_fasta_file_bzip2():
    sequence_file = SequenceFile(reference_file_bzip2)
    name, records = sequence_file.parse_sequence_file()

    assert 'genome' == name
    assert 1 == len(records)
    record = records[0]
    assert 'reference' == record.id
    assert 5180 == len(record)


def test_read_genbank_file():
    sequence_file = SequenceFile(reference_file_5000_snpeff)
    name, records = sequence_file.parse_sequence_file()

    assert 'NC_011083-5000' == name
    assert 1 == len(records)
    record = records[0]
    assert 'NC_011083.1' == record.id
    assert 5000 == len(record)


def test_records_fasta():
    sequence_file = SequenceFile(reference_file)
    records = list(sequence_file.records())

    assert 1 == len(records)
    record = records[0]
    assert 'reference' == record.id
    assert 5180 == len(record)


def test_records_genbank():
    sequence_file = SequenceFile(reference_file_5000_snpeff)
    records = list(sequence_file.records())

    assert 1 == len(records)
    record = records[0]
    assert 'NC_011083.1' == record.id
    assert 5000 == len(record)


def test_read_genbank_file_multiple_sequences():
    sequence_file = SequenceFile(reference_file_5000_snpeff_2)
    name, records = sequence_file.parse_sequence_file()

    assert 'NC_011083_CP001602-5000' == name
    assert 2 == len(records)

    record = records[0]
    assert 'NC_011083.1' == record.id
    assert 5000 == len(record)

    record = records[1]
    assert 'CP001602.2' == record.id
    assert 5000 == len(record)


def test_records_genbank_multiple_sequences():
    sequence_file = SequenceFile(reference_file_5000_snpeff_2)
    records = list(sequence_file.records())

    assert 2 == len(records)

    record = records[0]
    assert 'NC_011083.1' == record.id
    assert 5000 == len(record)

    record = records[1]
    assert 'CP001602.2' == record.id
    assert 5000 == len(record)


def test_write():
    with tempfile.TemporaryDirectory() as out_dir:
        out_file = Path(out_dir, 'output.fasta')
        sequence_file = SequenceFile(reference_file_5000_snpeff)
        sequence_file.write(out_file, 'fasta')

        assert out_file.exists()

        with open(out_file, 'r') as fh:
            sequences_out = list(SeqIO.parse(fh, 'fasta'))

        ref_name, sequences_in = sequence_file.parse_sequence_file()

        assert 1 == len(sequences_out)
        assert 1 == len(sequences_in)

        assert sequences_in[0].id == sequences_out[0].id
        assert sequences_in[0].seq == sequences_out[0].seq


def test_write_multiple_sequences():
    with tempfile.TemporaryDirectory() as out_dir:
        out_file = Path(out_dir, 'output.fasta')
        sequence_file = SequenceFile(reference_file_5000_snpeff_2)
        sequence_file.write(out_file, 'fasta')

        assert out_file.exists()

        with open(out_file, 'r') as fh:
            sequences_out = list(SeqIO.parse(fh, 'fasta'))

        ref_name, sequences_in = sequence_file.parse_sequence_file()

        assert 2 == len(sequences_out)
        assert 2 == len(sequences_in)

        assert sequences_in[0].id == sequences_out[0].id
        assert sequences_in[0].seq == sequences_out[0].seq
        assert sequences_in[1].id == sequences_out[1].id
        assert sequences_in[1].seq == sequences_out[1].seq


def test_create_snpeff_database():
    with tempfile.TemporaryDirectory() as out_dir:
        database_dir = Path(out_dir)
        snpeff_database_dir = database_dir / 'db'
        sequence_file = SequenceFile(reference_file_5000_snpeff)
        snpeff_database = sequence_file.create_snpeff_database(database_dir)

        expected_genbank_path = snpeff_database_dir / 'NC_011083-5000' / 'genes.gbk.gz'
        expected_snpeff_bin_file = snpeff_database_dir / 'NC_011083-5000' / 'snpEffectPredictor.bin'

        assert snpeff_database.config.name == 'snpEff.config'
        assert snpeff_database.config.exists()
        assert snpeff_database.config.parent == database_dir

        entries = parse_snpeff_config(snpeff_database.config)

        assert entries['data.dir'] == f'{snpeff_database_dir}'
        assert 'NC_011083-5000.genome' in entries
        assert entries['NC_011083-5000.chromosomes'] == 'NC_011083.1'
        assert entries['NC_011083-5000.NC_011083.1.codonTable'] == 'Standard'

        assert expected_genbank_path.exists()
        assert expected_genbank_path.is_symlink()
        assert expected_snpeff_bin_file.exists()


def test_can_use_snpeff():
    assert SequenceFile(reference_file_5000_snpeff).can_use_snpeff()
    assert SequenceFile(reference_file_5000_snpeff_2).can_use_snpeff()
    assert not SequenceFile(reference_file).can_use_snpeff()


def test_create_snpeff_database_multiple_contigs():
    with tempfile.TemporaryDirectory() as out_dir:
        database_dir = Path(out_dir)
        snpeff_database_dir = database_dir / 'db'
        sequence_file = SequenceFile(reference_file_5000_snpeff_2)

        snpeff_database = sequence_file.create_snpeff_database(database_dir, codon_type='Bacterial_and_Plant_Plastid')

        expected_genbank_path = snpeff_database_dir / 'NC_011083_CP001602-5000' / 'genes.gbk.gz'
        expected_snpeff_bin_file = snpeff_database_dir / 'NC_011083_CP001602-5000' / 'snpEffectPredictor.bin'

        assert snpeff_database.config.name == 'snpEff.config'
        assert snpeff_database.config.exists()
        assert snpeff_database.config.parent == database_dir

        entries = parse_snpeff_config(snpeff_database.config)

        assert entries['data.dir'] == f'{snpeff_database_dir}'
        assert 'NC_011083_CP001602-5000.genome' in entries
        assert entries['NC_011083_CP001602-5000.chromosomes'] == 'NC_011083.1,CP001602.2'
        assert entries['NC_011083_CP001602-5000.NC_011083.1.codonTable'] == 'Bacterial_and_Plant_Plastid'
        assert entries['NC_011083_CP001602-5000.CP001602.2.codonTable'] == 'Bacterial_and_Plant_Plastid'

        assert expected_genbank_path.exists()
        assert expected_genbank_path.is_symlink()
        assert expected_snpeff_bin_file.exists()



import logging
import shutil
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd

from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.index.KmerSearchManager import KmerSearchManagerSourmash
from genomics_data_index.storage.model.db import Sample, SampleKmerIndex
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service.SampleService import SampleService

logger = logging.getLogger(__name__)


class KmerService:
    FIND_MATCHES_MERGE_TYPES = ['union']

    def __init__(self, database_connection: DatabaseConnection, features_dir: Path, sample_service: SampleService):
        self._database = database_connection
        self._sample_service = sample_service
        self._features_dir = features_dir
        self._sourmash_search = KmerSearchManagerSourmash()

    def find_matches_within(self, sample_names: List[str],
                            kmer_size: int, distance_threshold: float,
                            results_merge_type: str = 'union',
                            samples_universe: SampleSet = None) -> SampleSet:
        """
        Find samples within a particular distance of a list of samples. This is based on kmer signatures/sketches.

        :param sample_names: The list of sample names to search for matches.
        :param kmer_size: The kmer size to use for searching through signatures/sketches.
        :param distance_threshold: A number from 0 to 1, with 0 indicating the distance threshold to include matches to
                                   to the samples listed in 'sample_names'.
        :param results_merge_type: Defines how to combine results when passing multiple 'sample_names'. A type of
                                   'union' means that matches will be the union of all samples matching anything in
                                   'sample_names'. Currently only 'union' is supported (this parameter is here to
                                   make a bit more clear how results are combined until I implement additional ways of
                                   merging results).
        :param samples_universe: The universe of samples to search through. Can be used to restrict which samples we will
                           consider for matches. Set to 'None' to search for matches in all samples in the system.
        :return: A SampleSet of the matches.
        """
        if results_merge_type != 'union':
            raise Exception(f'results_merge_type=[{results_merge_type}] is not supported. '
                            f'Only {self.FIND_MATCHES_MERGE_TYPES} are supported.')

        if samples_universe is None:
            sample_universe_objects = self._sample_service.get_samples()
        else:
            sample_universe_objects = self._sample_service.find_samples_by_ids(sample_ids=samples_universe)

        kmer_index_paths = [s.sample_kmer_index.kmer_index_path for s in sample_universe_objects if
                            s.sample_kmer_index is not None]

        if len(kmer_index_paths) < len(sample_universe_objects):
            logger.debug(f'Not all samples (number={len(sample_universe_objects)} have associated kmer signatures '
                         f'(number={len(kmer_index_paths)}). These will be excluded from the search.')

        if len(kmer_index_paths) == 0:
            return SampleSet.create_empty()
        else:
            similarity_threshold = 1 - distance_threshold

            matches_df = pd.DataFrame(data=[], columns=[
                'Query',
                'Match',
                'Similarity',
                'Distance',
            ])

            for sample_name in sample_names:
                query_sample = self._sample_service.get_sample(sample_name)
                results_df = self._sourmash_search.search(kmer_size=kmer_size,
                                                          similarity_threshold=similarity_threshold,
                                                          query_file=query_sample.sample_kmer_index.kmer_index_path,
                                                          search_files=kmer_index_paths)
                results_df['Distance'] = 1 - results_df['similarity']
                results_df['Query'] = sample_name
                results_df = results_df.rename({
                    'name': 'Match',
                    'similarity': 'Similarity',
                }, axis='columns')
                results_df = results_df[['Query', 'Match', 'Similarity', 'Distance']]

                matches_df = pd.concat([matches_df, results_df])

            sample_name_ids = self._sample_service.find_sample_name_ids(set(matches_df['Match'].tolist()))
            matches_set = SampleSet(sample_name_ids.values())

            return matches_set

    def get_distance_matrix(self, sample_ids: Union[List[int], SampleSet], kmer_size: int,
                            ncores: int = 1) -> Tuple[
        np.ndarray, List[str]]:
        if isinstance(sample_ids, list):
            sample_ids = SampleSet(sample_ids)

        sourmash_search_multicore = KmerSearchManagerSourmash(ncores=ncores)

        samples = self._sample_service.find_samples_by_ids(sample_ids)
        kmer_index_paths = [s.sample_kmer_index.kmer_index_path for s in samples if
                            s.sample_kmer_index is not None]

        if len(kmer_index_paths) < len(samples):
            raise Exception(f'Not all samples (number={len(samples)}) have associated kmer signatures '
                            f'(number={len(kmer_index_paths)}).')

        return sourmash_search_multicore.distances(kmer_size=kmer_size, signature_files=kmer_index_paths)

    def has_kmer_index(self, sample_name: str) -> bool:
        if self._sample_service.exists(sample_name):
            sample = self._sample_service.get_sample(sample_name)
            return sample.sample_kmer_index is not None
        else:
            return False

    def insert_kmer_index(self, sample_name: str, kmer_index_path: Path):

        if self._sample_service.exists(sample_name):
            sample = self._sample_service.get_sample(sample_name)
        else:
            sample = Sample(name=sample_name)
            self._database.get_session().add(sample)

        kmer_path_internal = self._features_dir / kmer_index_path.name
        shutil.copy(kmer_index_path, kmer_path_internal)
        kmer_index = SampleKmerIndex(sample=sample, kmer_index_path=kmer_path_internal)
        self._database.get_session().add(kmer_index)
        self._database.get_session().commit()



import warnings
from pathlib import Path
from tempfile import TemporaryDirectory

warnings.filterwarnings("ignore", category=DeprecationWarning)

from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.test.integration import sample_dirs, reference_file
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor


def test_connection():
    with TemporaryDirectory() as tmp_dir:
        connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                 database_dir=Path(tmp_dir))

        assert connection.sample_service is not None
        assert connection.variation_service is not None
        assert connection.reference_service is not None


def test_add_reference_from_connection():
    with TemporaryDirectory() as tmp_dir:
        connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                 database_dir=Path(tmp_dir))

        connection.reference_service.add_reference_genome(reference_file)


def test_add_variants_from_connection():
    with TemporaryDirectory() as tmp_dir:
        connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                 database_dir=Path(tmp_dir))

        data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                      sample_files_processor=SerialSampleFilesProcessor(
                                                                          Path(tmp_dir))
                                                                      )

        connection.reference_service.add_reference_genome(reference_file)
        connection.variation_service.insert(feature_scope_name='genome',
                                            data_package=data_package)



from __future__ import annotations

import logging
from pathlib import Path

import pandas as pd

from genomics_data_index.configuration.connector.FilesystemStorage import FilesystemStorage
from genomics_data_index.storage.model.db.DatabasePathTranslator import DatabasePathTranslator
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service.CoreAlignmentService import CoreAlignmentService
from genomics_data_index.storage.service.KmerService import KmerService
from genomics_data_index.storage.service.MLSTService import MLSTService
from genomics_data_index.storage.service.ReferenceService import ReferenceService
from genomics_data_index.storage.service.SampleService import SampleService
from genomics_data_index.storage.service.TreeService import TreeService
from genomics_data_index.storage.service.VariationService import VariationService

logger = logging.getLogger(__name__)


class DataIndexConnection:

    def __init__(self, reference_service: ReferenceService, sample_service: SampleService,
                 variation_service: VariationService, alignment_service: CoreAlignmentService,
                 tree_service: TreeService, kmer_service: KmerService,
                 mlst_service: MLSTService, filesystem_storage: FilesystemStorage,
                 database_connection: DatabaseConnection):
        self._reference_service = reference_service
        self._sample_service = sample_service
        self._variation_service = variation_service
        self._alignment_service = alignment_service
        self._tree_service = tree_service
        self._kmer_service = kmer_service
        self._mlst_service = mlst_service
        self._filesystem_storage = filesystem_storage
        self._database_connection = database_connection

    @property
    def database(self):
        return self._database_connection

    @property
    def reference_service(self):
        return self._reference_service

    @property
    def sample_service(self):
        return self._sample_service

    @property
    def variation_service(self):
        return self._variation_service

    @property
    def alignment_service(self):
        return self._alignment_service

    @property
    def tree_service(self):
        return self._tree_service

    @property
    def kmer_service(self):
        return self._kmer_service

    @property
    def mlst_service(self):
        return self._mlst_service

    @property
    def filesystem_storage(self):
        return self._filesystem_storage

    def db_size(self, unit: str = 'B') -> pd.DataFrame:
        if unit == 'B':
            factor = 1
        elif unit == 'KB':
            factor = 1024
        elif unit == 'MB':
            factor = 1024 ** 2
        elif unit == 'GB':
            factor = 1024 ** 3
        else:
            raise Exception(f'Unknown unit=[{unit}]')

        filesystem_df = self.filesystem_storage.get_storage_size()
        database_df = self._database_connection.get_database_size()

        size_df = pd.concat([filesystem_df, database_df])
        total_data_size = size_df['Data Size'].sum()
        total_index_size = size_df['Index Size'].sum()
        total_items = size_df['Number of Items'].sum()
        total_row = pd.DataFrame([['Total', pd.NA, pd.NA, total_data_size, total_index_size, total_items]],
                                 columns=['Type', 'Name', 'Division', 'Data Size', 'Index Size', 'Number of Items'])
        size_df = pd.concat([size_df, total_row])

        # Reorder columns
        size_df = size_df[['Type', 'Name', 'Division', 'Data Size', 'Index Size', 'Number of Items']]

        size_df['Data Size'] = size_df['Data Size'] / factor
        size_df['Index Size'] = size_df['Index Size'] / factor
        size_df = size_df.rename({'Data Size': f'Data Size ({unit})',
                                  'Index Size': f'Index Size ({unit})'}, axis='columns')
        return size_df

    @classmethod
    def connect(cls, database_connection: str, database_dir: Path,
                index_unknown_missing: bool = True) -> DataIndexConnection:
        filesystem_storage = FilesystemStorage(Path(database_dir))
        dpt = DatabasePathTranslator(filesystem_storage.root_dir)
        logger.debug(f'Using database directory {database_dir}')

        logger.debug(f'Connecting to database {database_connection}')
        database = DatabaseConnection(connection_string=database_connection,
                                      database_path_translator=dpt)

        # According to <https://www.sqlite.org/limits.html> (**Maximum Number Of Host Parameters In A Single
        #  SQL Statement**), the maximum number of parameters is 32766 (for SQLite, but depends on version).
        # TODO: Make this configurable
        sql_select_limit = 32766

        reference_service = ReferenceService(database, filesystem_storage.reference_dir)
        sample_service = SampleService(database, sql_select_limit=sql_select_limit)
        variation_service = VariationService(database_connection=database,
                                             variation_dir=filesystem_storage.variation_dir,
                                             reference_service=reference_service,
                                             sample_service=sample_service,
                                             index_unknown_missing=index_unknown_missing,
                                             sql_select_limit=sql_select_limit)

        alignment_service = CoreAlignmentService(database=database,
                                                 reference_service=reference_service,
                                                 sample_service=sample_service,
                                                 variation_service=variation_service)

        tree_service = TreeService(database, reference_service, alignment_service)

        kmer_service = KmerService(database_connection=database,
                                   sample_service=sample_service,
                                   features_dir=filesystem_storage.kmer_dir)

        mlst_service = MLSTService(database_connection=database, sample_service=sample_service,
                                   mlst_dir=filesystem_storage.mlst_dir)

        return DataIndexConnection(reference_service=reference_service, sample_service=sample_service,
                                   variation_service=variation_service, alignment_service=alignment_service,
                                   tree_service=tree_service, kmer_service=kmer_service, mlst_service=mlst_service,
                                   filesystem_storage=filesystem_storage, database_connection=database)



import logging
from typing import Union, Iterable, Tuple, List

from ete3 import Tree

from genomics_data_index.api.query.TreeBuilder import TreeBuilder
from genomics_data_index.configuration.connector import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet

logger = logging.getLogger(__name__)


class TreeBuilderReferenceMutations(TreeBuilder):
    TREE_METHODS = ['iqtree']

    def __init__(self, database_connection: DataIndexConnection, reference_name: str):
        super().__init__()
        self._database_connection = database_connection
        self._reference_name = reference_name

        if not database_connection.reference_service.exists_reference_genome(reference_name):
            raise Exception(f'Reference genome [{reference_name}] does not exist.')

    def build(self, samples_set: Union[SampleSet, Iterable[str]],
              method: str = 'iqtree', include_reference=True, **kwargs) -> Tuple[Tree, int, SampleSet]:
        if method == 'iqtree':
            return self._build_iqtree(samples_set=samples_set, include_reference=include_reference,
                                      **kwargs)
        else:
            raise Exception(f'Invalid method={method}, only support: {self.TREE_METHODS}')

    def _build_iqtree(self, samples_set: Union[SampleSet, Iterable[str]],
                      ncores: int = 1,
                      align_type: str = 'full',
                      include_variants: List[str] = None,
                      include_reference: bool = True,
                      extra_params: str = None) -> Tuple[Tree, int, SampleSet]:
        sample_service = self._database_connection.sample_service
        alignment_service = self._database_connection.alignment_service
        tree_service = self._database_connection.tree_service

        if isinstance(samples_set, SampleSet):
            samples_names = {s.name for s in sample_service.find_samples_by_ids(samples_set)}
        else:
            samples_names = set(samples_set)

        reference_samples = {s.name for s in sample_service.get_samples_associated_with_reference(self._reference_name)}

        tree_samples = samples_names.intersection(reference_samples)
        tree_samples_set = SampleSet(sample_service.find_sample_name_ids(tree_samples).values())

        if len(tree_samples) < len(samples_names):
            logger.warning(f'Building build_tree with only ({len(tree_samples)}/{len(reference_samples)}) samples '
                           f'that are found on reference [{self._reference_name}]')

        logger.info(f'Building build_tree using "iqtree" for {len(tree_samples_set)} samples')
        alignment_data = alignment_service.construct_alignment(reference_name=self._reference_name,
                                                               samples=samples_names,
                                                               align_type=align_type,
                                                               include_variants=include_variants,
                                                               include_reference=include_reference)
        tree_data, out = tree_service.build_tree(alignment_data, tree_build_type='iqtree',
                                                 num_cores=ncores, align_type=align_type,
                                                 extra_params=extra_params)

        return tree_data, alignment_data.get_alignment_length(), tree_samples_set



import math

import pytest

from genomics_data_index.storage.SampleSet import SampleSet


def test_create_sample_set_from_list():
    sample_set = SampleSet(sample_ids=[1])

    assert 1 in sample_set
    assert 2 not in sample_set
    assert 1 == len(sample_set)
    assert {1} == set(sample_set)
    assert not sample_set.is_empty()


def test_create_sample_set_from_set():
    assert {1} == set(SampleSet(sample_ids={1}))
    assert {1, 10} == set(SampleSet(sample_ids={1, 10}))
    assert set() == set(SampleSet(sample_ids=set()))


def test_create_sample_set_from_list_multiple():
    sample_set = SampleSet(sample_ids=[1, 10])

    assert 1 in sample_set
    assert 2 not in sample_set
    assert 10 in sample_set
    assert 11 not in sample_set
    assert 2 == len(sample_set)
    assert {1, 10} == set(sample_set)


def test_create_sample_set_from_list_empty():
    sample_set = SampleSet(sample_ids=[])

    assert 1 not in sample_set
    assert 0 == len(sample_set)
    assert set() == set(sample_set)
    assert sample_set.is_empty()


def test_serialize_deserialize():
    sample_set = SampleSet(sample_ids=[1, 3, 10])

    assert {1, 3, 10} == set(sample_set)

    sample_set_bytes = sample_set.get_bytes()
    assert isinstance(sample_set_bytes, bytes)

    sample_set_deserialize = SampleSet.from_bytes(sample_set_bytes)
    assert {1, 3, 10} == set(sample_set_deserialize)


def test_intersect_sample_set():
    sample_set1 = SampleSet(sample_ids=[1, 3, 10])
    sample_set2 = SampleSet(sample_ids=[3, 10, 20])
    sample_set_empty = SampleSet(sample_ids=[])
    sample_set_non_overlap = SampleSet(sample_ids=[50, 100])

    intersection = sample_set1.intersection(sample_set2)
    assert isinstance(intersection, SampleSet)
    assert {3, 10} == set(intersection)

    intersection = sample_set2.intersection(sample_set1)
    assert isinstance(intersection, SampleSet)
    assert {3, 10} == set(intersection)

    intersection = sample_set1.intersection(sample_set_empty)
    assert isinstance(intersection, SampleSet)
    assert set() == set(intersection)

    intersection = sample_set1.intersection(sample_set_non_overlap)
    assert isinstance(intersection, SampleSet)
    assert set() == set(intersection)


def test_union_sample_set():
    sample_set1 = SampleSet(sample_ids=[1, 3, 10])
    sample_set2 = SampleSet(sample_ids=[3, 10, 20])
    sample_set_empty = SampleSet(sample_ids=[])
    sample_set_non_overlap = SampleSet(sample_ids=[50, 100])

    union = sample_set1.union(sample_set2)
    assert isinstance(union, SampleSet)
    assert {1, 3, 10, 20} == set(union)

    union = sample_set2.union(sample_set1)
    assert isinstance(union, SampleSet)
    assert {1, 3, 10, 20} == set(union)

    union = sample_set1.union(sample_set_empty)
    assert isinstance(union, SampleSet)
    assert {1, 3, 10} == set(union)

    union = sample_set1.union(sample_set_non_overlap)
    assert isinstance(union, SampleSet)
    assert {1, 3, 10, 50, 100} == set(union)


def test_jaccard_index_sample_set():
    sample_set1 = SampleSet(sample_ids=[1, 3, 10])
    sample_set2 = SampleSet(sample_ids=[3, 10, 20])
    sample_set_empty = SampleSet(sample_ids=[])
    sample_set_non_overlap = SampleSet(sample_ids=[50, 100])
    sample_set_3 = SampleSet(sample_ids=[1, 3, 10, 20, 30])

    assert math.isclose(1.0, sample_set1.jaccard_index(sample_set1), rel_tol=1e-3)

    assert math.isclose(0.5, sample_set1.jaccard_index(sample_set2), rel_tol=1e-3)
    assert math.isclose(0.5, sample_set2.jaccard_index(sample_set1), rel_tol=1e-3)

    assert math.isclose(0, sample_set1.jaccard_index(sample_set_empty), rel_tol=1e-3)
    assert math.isclose(0, sample_set_empty.jaccard_index(sample_set1), rel_tol=1e-3)

    assert math.isclose(0, sample_set1.jaccard_index(sample_set_non_overlap), rel_tol=1e-3)
    assert math.isclose(0, sample_set_non_overlap.jaccard_index(sample_set1), rel_tol=1e-3)

    assert math.isclose(0.6, sample_set1.jaccard_index(sample_set_3), rel_tol=1e-3)
    assert math.isclose(0.6, sample_set_3.jaccard_index(sample_set1), rel_tol=1e-3)


def test_jaccard_index_sample_ids_set():
    sample_set1 = SampleSet(sample_ids=[1, 3, 10])
    sample_set2 = {3, 10, 20}
    sample_set_empty = set()
    sample_set_non_overlap = {50, 100}
    sample_set_3 = {1, 3, 10, 20, 30}

    assert math.isclose(1.0, sample_set1.jaccard_index(sample_set1), rel_tol=1e-3)
    assert math.isclose(0.5, sample_set1.jaccard_index(sample_set2), rel_tol=1e-3)
    assert math.isclose(0, sample_set1.jaccard_index(sample_set_empty), rel_tol=1e-3)
    assert math.isclose(0, sample_set1.jaccard_index(sample_set_non_overlap), rel_tol=1e-3)
    assert math.isclose(0.6, sample_set1.jaccard_index(sample_set_3), rel_tol=1e-3)


def test_minus_sample_set():
    set1 = SampleSet(sample_ids=[1, 3, 10])
    set2 = SampleSet(sample_ids=[3, 10, 20])
    set3 = SampleSet(sample_ids=[40, 60, 80, 100])

    assert {1} == set(set1.minus(set2))
    assert {20} == set(set2.minus(set1))
    assert set() == set(set1.minus(set1))
    assert {1, 3, 10} == set(set1.minus(set3))
    assert {40, 60, 80, 100} == set(set3.minus(set1))
    assert {1, 3, 10} == set(set1.minus(SampleSet.create_empty()))


def test_intersect_python_set():
    sample_set1 = SampleSet(sample_ids=[1, 3, 10])

    assert {3, 10} == set(sample_set1.intersection({3, 10, 20}))


def test_union_python_set():
    sample_set1 = SampleSet(sample_ids=[1, 3, 10])

    assert {1, 3, 10, 20} == set(sample_set1.union({3, 10, 20}))
    assert {1, 3, 10} == set(sample_set1.union(set()))
    assert {1, 3, 10, 50, 100} == set(sample_set1.union({50, 100}))


def test_create_empty_sample_set():
    empty_set = SampleSet.create_empty()
    assert len(empty_set) == 0
    assert set() == set(empty_set)
    assert empty_set.is_empty()


def test_create_all_sample_set():
    all_set = SampleSet.create_all()
    assert len(all_set) == 2 ** 32
    assert 0 in all_set
    assert 1 in all_set
    assert 2 ** 32 - 1 in all_set
    assert 2 ** 32 not in all_set

    other_set = SampleSet(sample_ids=[1, 3, 10])
    empty_set = SampleSet.create_empty()

    assert all_set.intersection(other_set) == other_set
    assert other_set.intersection(all_set) == other_set
    assert isinstance(all_set.intersection(other_set), SampleSet)
    assert isinstance(other_set.intersection(all_set), SampleSet)

    assert all_set.union(other_set) == all_set
    assert other_set.union(all_set) == all_set
    assert all_set.union(all_set) == all_set

    assert empty_set.intersection(all_set).is_empty()
    assert all_set.intersection(empty_set).is_empty()
    assert empty_set.union(all_set) == all_set
    assert all_set.union(empty_set) == all_set


def test_minus_2():
    sample_set1 = SampleSet(sample_ids=[1, 3, 10])
    sample_set1_identical = SampleSet(sample_ids=[1, 3, 10])
    sample_set2 = SampleSet(sample_ids=[3, 10, 20])
    sample_set3 = SampleSet(sample_ids=[3])
    sample_set4 = SampleSet(sample_ids=[1])
    sample_set_empty = SampleSet(sample_ids=[])
    sample_set_non_overlap = SampleSet(sample_ids=[50, 100])

    subtract = sample_set1.minus(sample_set2)
    assert isinstance(subtract, SampleSet)
    assert {1} == set(subtract)

    subtract = sample_set2.minus(sample_set1)
    assert isinstance(subtract, SampleSet)
    assert {20} == set(subtract)

    subtract = sample_set1.minus(sample_set_empty)
    assert isinstance(subtract, SampleSet)
    assert {1, 3, 10} == set(subtract)

    subtract = sample_set_empty.minus(sample_set1)
    assert isinstance(subtract, SampleSet)
    assert set() == set(subtract)

    subtract = sample_set1.minus(sample_set_non_overlap)
    assert isinstance(subtract, SampleSet)
    assert {1, 3, 10} == set(subtract)

    subtract = sample_set1.minus(sample_set1_identical)
    assert isinstance(subtract, SampleSet)
    assert set() == set(subtract)

    subtract = sample_set1.minus(sample_set3)
    assert isinstance(subtract, SampleSet)
    assert {1, 10} == set(subtract)

    subtract = sample_set1.minus([sample_set3, sample_set4])
    assert isinstance(subtract, SampleSet)
    assert {10} == set(subtract)

    # Pass in a set of integers
    subtract = sample_set1.minus({1, 10})
    assert isinstance(subtract, SampleSet)
    assert {3} == set(subtract)

    # Test case of all set
    all_set = SampleSet.create_all()
    subtract = sample_set1.minus(all_set)
    assert isinstance(subtract, SampleSet)
    assert set() == set(subtract)

    # Test case of all set other
    with pytest.raises(Exception) as execinfo:
        all_set.minus(sample_set1)
    assert 'Cannot subtract anything from AllSampleSet' in str(execinfo.value)



import abc
from typing import Union

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.storage.SampleSet import SampleSet


class ClusterScoreMethod(abc.ABC):
    """
    Defines a method for assigning scores to clusters of samples based on their relationship in a tree.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def score(self, data: Union[SamplesQuery, SampleSet], universe_samples: SamplesQuery) -> float:
        """
        Scores the passed set of samples (within the given universe of samples). The score is a number defining
        how well these samples cluster together in a tree. An implementation of ClusterScoreMethod can be regisered
        in :py:class:`genomics_data_index.api.query.impl.ClusterScorer` and then executed with ClusterScorer.score_samples().
        :param data: The data to score (set of samples).
        :param universe_samples: The set of universe samples.
        :return: A float defining the score of the passed set of samples in a tree.
        """
        pass



from pathlib import Path


class DatabasePathTranslator:
    """
    Used to translate between absolute and relative paths when persisting Paths in the database.
    """

    def __init__(self, root_path):
        self._root_path = root_path

    def to_database(self, file: Path) -> str:
        """
        Translates the given file to the relative representation to store in the database.

        :param file: The file to translate.
        :return: A string with the relative representation to store in the database.
        """
        try:
            relative_path = file.relative_to(self._root_path)
            return str(relative_path)
        except ValueError as e:
            raise Exception(f'File to save [{file}] is not relative to root data path [{self._root_path}]', e)

    def from_database(self, file: str) -> Path:
        """
        Translates the given file representation in the database to an absolute Path on the filesystem.

        :param file: The string representation of the file in the database.
        :return: A Path representing this file on the filesystem.
        """
        file_path = Path(file)
        if file_path.is_absolute():
            raise Exception(f'Path from database [{file_path}] is absolute')

        absolute_file_path = self._root_path / file_path

        if not absolute_file_path.exists():
            raise Exception(f'absolute_file_path=[{absolute_file_path}] does not exist for relative path [{file_path}]')

        return absolute_file_path



import copy
import logging
import tempfile
import time
from pathlib import Path
from typing import List, Dict, Generator

from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.mutation.VariationFile import VariationFile
from genomics_data_index.storage.model.db import SampleNucleotideVariation
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service.ReferenceService import ReferenceService
from genomics_data_index.storage.service.SampleService import SampleService
from genomics_data_index.storage.service.VariationService import VariationService

logger = logging.getLogger(__name__)


class CoreAlignmentService:
    ALIGN_TYPES = ['core', 'full']
    INCLUDE_VARIANT_SUBSTITUTION = ['SNP', 'MNP']
    INCLUDE_VARIANT_DELETION = ['DELETION', 'DELETION_OTHER']
    INCLUDE_VARIANT_TYPES = INCLUDE_VARIANT_SUBSTITUTION + INCLUDE_VARIANT_DELETION
    INCLUDE_VARIANT_DEFAULT = ['SNP', 'MNP', 'DELETION']

    # Mask generated sequence with this character (which should not appear anywhere else) so I can remove
    # all positions with this character later (when generating core alignment)
    CORE_MASK_CHAR = '?'

    def __init__(self, database: DatabaseConnection, reference_service: ReferenceService,
                 sample_service: SampleService, variation_service: VariationService):
        self._database = database
        self._reference_service = reference_service
        self._variation_service = variation_service
        self._sample_service = sample_service

    def _all_sample_names(self, reference_name: str) -> List[str]:
        samples = self._sample_service.get_samples_with_variants(reference_name)
        return [s.name for s in samples]

    def _create_core_mask(self, sample_variations: List[SampleNucleotideVariation]) -> MaskedGenomicRegions:
        masked_regions = [v.masked_regions for v in sample_variations]
        return MaskedGenomicRegions.union_all(masked_regions)

    def _get_core_positions(self, sample_variations: List[SampleNucleotideVariation],
                            core_mask: MaskedGenomicRegions, include_expression: str) -> Dict[str, List[int]]:
        variation_files = [v.nucleotide_variants_file for v in sample_variations]
        union_df = VariationFile.union_all_files(variation_files, include_expression=include_expression)
        union_df = union_df.sort_values(['CHROM', 'POS'])
        core_positions = {}
        for index, row in union_df.iterrows():
            seq_name = row['CHROM']
            position = row['POS']
            if not core_mask.contains(seq_name, position, start_position_index='1'):
                if seq_name not in core_positions:
                    core_positions[seq_name] = [position]
                else:
                    core_positions[seq_name].append(position)

        return core_positions

    def _extract_sequence_from_positions(self, input_seq: Seq, positions: List[int]) -> Seq:
        sequence_string = ''

        for position in positions:
            sequence_string += str(input_seq[position - 1:position])

        return Seq(sequence_string)

    def _core_alignment_sequence_generator(self, reference_file: Path, core_positions: Dict[str, List[int]],
                                           sample_variations: List[SampleNucleotideVariation],
                                           include_expression: str) -> Generator[
        Dict[str, SeqRecord], None, None]:
        for sample_variation in sample_variations:
            seq_records = {}

            consensus_records = VariationFile(sample_variation.nucleotide_variants_file).consensus(
                reference_file, include_expression=include_expression)
            for record in consensus_records:
                sequence_name = record.id
                record.id = sample_variation.sample.name
                record.description = 'generated automatically'
                record.seq = self._extract_sequence_from_positions(input_seq=record.seq,
                                                                   positions=core_positions[sequence_name])

                seq_records[sequence_name] = record
            yield seq_records

    def _full_alignment_sequence_generator(self, reference_file: Path,
                                           sample_variations: List[SampleNucleotideVariation],
                                           include_expression: str) -> Generator[
        Dict[str, SeqRecord], None, None]:
        for sample_variation in sample_variations:
            seq_records = {}

            consensus_records = VariationFile(sample_variation.nucleotide_variants_file).consensus(
                reference_file=reference_file,
                include_expression=include_expression,
                mask_file=sample_variation.masked_regions_file
            )
            for record in consensus_records:
                seq_records[record.id] = record
                record.id = sample_variation.sample.name
                record.description = 'generated automatically'
            yield seq_records

    def construct_alignment(self, reference_name: str, samples: List[str] = None,
                            include_reference: bool = True, align_type: str = 'core',
                            include_variants: List[str] = None) -> MultipleSeqAlignment:
        if samples is None or len(samples) == 0:
            samples = self._all_sample_names(reference_name)

        if include_variants is None:
            include_variants = self.INCLUDE_VARIANT_DEFAULT

        if align_type == 'core' and include_variants != ['SNP']:
            raise Exception(f'align_type={align_type} and include_variants={include_variants}. Currently'
                            f' align_type=core only works with include_variants=["SNP"]')
        else:
            subtitution_types = []
            deletion_types = []
            for variant_type in include_variants:
                if variant_type not in self.INCLUDE_VARIANT_TYPES:
                    raise Exception(f'variant_type={variant_type} found in include_variants={include_variants}. '
                                    f'Only {self.INCLUDE_VARIANT_TYPES} are supported')
                elif variant_type in self.INCLUDE_VARIANT_SUBSTITUTION:
                    subtitution_types.append(variant_type)
                elif variant_type in self.INCLUDE_VARIANT_DELETION:
                    deletion_types.append(variant_type)

        include_expression = ''
        if len(deletion_types) == 1:
            deletion_type = deletion_types[0]
            # To make sure I only include deletions (out of all INDELs) I use (ILEN<0) which
            # represents INDELs that are deletions (see <http://samtools.github.io/bcftools/bcftools.html#expressions>).
            if deletion_type == 'DELETION':
                include_expression = '((INFO/TYPE=="INDEL") & (ILEN<0))'
            elif deletion_type == 'DELETION_OTHER':
                include_expression = '(ILEN<0)'
            else:
                raise Exception(f'deletion_type={deletion_type} is invalid. Must be one of '
                                f'{self.INCLUDE_VARIANT_DELETION}')
        elif len(deletion_types) > 1:
            raise Exception(f'Can only set one of {self.INCLUDE_VARIANT_DELETION}, got {deletion_types}')

        if len(subtitution_types) > 0 and len(include_expression) > 0:
            include_expression = include_expression + ' | '

        include_expression = include_expression + ' | '.join(
            f'(INFO/TYPE=="{variant}")' for variant in subtitution_types)

        sample_nucleotide_variants = self._variation_service.get_sample_nucleotide_variation(samples)

        alignment_seqs = {}
        alignments = {}

        start_time = time.time()
        logger.info(f'Started building alignment for {len(samples)} samples with include_variants={include_variants}')
        logger.debug(f'Alignment include_expression=\'{include_expression}\'')

        with tempfile.TemporaryDirectory() as tmp_dir:
            # Write reference genome to file for creating consensus sequences
            reference_file = Path(tmp_dir) / 'reference.fasta'
            reference_records = self._reference_service.get_reference_genome_records(reference_name)
            with open(reference_file, 'w') as h:
                SeqIO.write(reference_records, h, 'fasta')

            if align_type == 'core':
                core_mask = self._create_core_mask(sample_nucleotide_variants)
                core_positions = self._get_core_positions(sample_variations=sample_nucleotide_variants,
                                                          core_mask=core_mask,
                                                          include_expression=include_expression)

                alignment_generator = self._core_alignment_sequence_generator(
                    reference_file=reference_file,
                    sample_variations=sample_nucleotide_variants,
                    core_positions=core_positions,
                    include_expression=include_expression
                )

                if include_reference:
                    for record in reference_records:
                        sequence_name = record.id
                        core_seq = self._extract_sequence_from_positions(input_seq=record.seq,
                                                                         positions=core_positions[sequence_name])
                        new_record = SeqRecord(id=reference_name,
                                               seq=core_seq,
                                               description='[reference genome]')
                        alignment_seqs[sequence_name] = [new_record]

            elif align_type == 'full':
                alignment_generator = self._full_alignment_sequence_generator(
                    reference_file=reference_file,
                    sample_variations=sample_nucleotide_variants,
                    include_expression=include_expression
                )

                if include_reference:
                    # Add the reference sequence in
                    reference_records_align = copy.deepcopy(reference_records)
                    for record in reference_records_align:
                        alignment_seqs[record.id] = [record]
                        record.id = reference_name
                        record.description = '[reference genome]'
            else:
                raise Exception(f'Unknown value for align_type=[{align_type}]. Must be one of {self.ALIGN_TYPES}')

            # Actually generate the data
            for sequence_record in alignment_generator:
                for sequence in sequence_record:
                    if sequence == reference_name:
                        raise Exception(
                            f'Data contains a sequence with name [{sequence}], which is the same as the reference genome name')
                    elif sequence not in alignment_seqs:
                        alignment_seqs[sequence] = [sequence_record[sequence]]
                    else:
                        alignment_seqs[sequence].append(sequence_record[sequence])

        for sequence_name in alignment_seqs:
            alignments[sequence_name] = MultipleSeqAlignment(
                alignment_seqs[sequence_name])
            alignments[sequence_name].sort()

        sequence_names = sorted(alignments.keys())
        seq1 = sequence_names.pop()
        snv_align = alignments[seq1]

        for sequence_name in sequence_names:
            snv_align += alignments[sequence_name]

        end_time = time.time()
        logger.info(f'Finished building alignment for {len(samples)} samples. '
                    f'Took {end_time - start_time:0.2f} seconds')

        return snv_align



import re
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import cast, Dict

import pytest
from pybedtools import BedTool

from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.mutation.variants_processor.MultipleProcessVcfVariantsTableProcessor import \
    MultipleProcessVcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.mutation.variants_processor.SerialVcfVariantsTableProcessor import \
    SerialVcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.processor.MultipleProcessSampleFilesProcessor import \
    MultipleProcessSampleFilesProcessor
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor
from genomics_data_index.test.integration import data_dir
from genomics_data_index.test.integration.storage.io.mutation import vcf_and_mask_files, vcf_and_bed_mask_files


def vcf_and_mixed_mask_files() -> Dict[str, Dict[str, Path]]:
    sample_vcf_map = {}
    sample_genomic_files_mask = {}

    sample_vcf_map['SampleA'] = data_dir / 'SampleA' / 'snps.vcf.gz'
    sample_genomic_files_mask['SampleA'] = data_dir / 'SampleA' / 'snps.aligned.bed.gz'
    sample_vcf_map['SampleB'] = data_dir / 'SampleB' / 'snps.vcf.gz'
    sample_genomic_files_mask['SampleB'] = data_dir / 'SampleB' / 'snps.aligned.fa'
    sample_vcf_map['SampleC'] = data_dir / 'SampleC' / 'snps.vcf.gz'
    sample_genomic_files_mask['SampleC'] = None

    return {
        'vcfs': sample_vcf_map,
        'masks': sample_genomic_files_mask
    }


def test_iter_sample_data(sample_dirs):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        vcf_masks = vcf_and_mask_files(sample_dirs)
        file_processor = SerialSampleFilesProcessor(tmp_file)
        data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                         masked_genomic_files_map=vcf_masks[
                                                                             'masks'],
                                                                         sample_files_processor=file_processor)

        processed_files_dict = {}
        for sample_file in data_package.iter_sample_data():
            processed_files_dict[sample_file.sample_name] = sample_file

        assert 3 == len(processed_files_dict)

        count = 0
        for sample, mask_len in [('SampleA', 437), ('SampleB', 276), ('SampleC', 329)]:
            sample_data = processed_files_dict[sample]
            assert isinstance(sample_data, NucleotideSampleData)
            sample_data = cast(NucleotideSampleData, sample_data)
            mask = sample_data.get_mask()
            assert mask_len == len(mask)
            assert {'reference'} == mask.sequence_names()
            mask_file = sample_data.get_mask_file()
            assert mask_file.parent == tmp_file
            vcf_file, vcf_index = sample_data.get_vcf_file()
            assert vcf_file.exists()
            assert vcf_index.exists()
            assert vcf_file.parent == tmp_file
            assert vcf_index.parent == tmp_file

            count += 1

        # Make sure I tested all 3 files in above loop
        assert 3 == count


def test_iter_sample_data_snippy(sample_dirs):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        file_processor = SerialSampleFilesProcessor(tmp_file)
        expected_data = vcf_and_bed_mask_files(sample_dirs)
        data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                      sample_files_processor=file_processor)

        processed_files_dict = {}
        for sample_file in data_package.iter_sample_data():
            processed_files_dict[sample_file.sample_name] = sample_file

        assert 3 == len(processed_files_dict)

        count = 0
        for sample in ['SampleA', 'SampleB', 'SampleC']:
            sample_data = processed_files_dict[sample]
            assert isinstance(sample_data, NucleotideSampleData)
            sample_data = cast(NucleotideSampleData, sample_data)
            mask = sample_data.get_mask()

            expected_mask = BedTool(str(expected_data['masks_minus_vcf'][sample]))
            assert expected_mask == mask.mask, f'expected=\n{expected_mask} != actual=\n{mask.mask}'

            assert {'reference'} == mask.sequence_names()
            mask_file = sample_data.get_mask_file()
            assert mask_file.parent == tmp_file
            vcf_file, vcf_index = sample_data.get_vcf_file()
            assert vcf_file.exists()
            assert vcf_index.exists()
            assert vcf_file.parent == tmp_file
            assert vcf_index.parent == tmp_file

            count += 1

        # Make sure I tested all 3 files in above loop
        assert 3 == count


def test_iter_sample_data_bed_masks(sample_dirs):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        vcf_masks = vcf_and_bed_mask_files(sample_dirs)
        file_processor = SerialSampleFilesProcessor(tmp_file)
        data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                         masked_genomic_files_map=vcf_masks[
                                                                             'masks'],
                                                                         sample_files_processor=file_processor)

        processed_files_dict = {}
        for sample_file in data_package.iter_sample_data():
            processed_files_dict[sample_file.sample_name] = sample_file

        assert 3 == len(processed_files_dict)
        assert {'SampleA', 'SampleB', 'SampleC'} == set(processed_files_dict.keys())

        count = 0
        for sample, mask_len in [('SampleA', 437), ('SampleB', 276), ('SampleC', 329)]:
            sample_data = processed_files_dict[sample]
            assert isinstance(sample_data, NucleotideSampleData)
            sample_data = cast(NucleotideSampleData, sample_data)
            mask = sample_data.get_mask()
            assert mask_len == len(mask)
            assert {'reference'} == mask.sequence_names()
            mask_file = sample_data.get_mask_file()
            assert mask_file.parent == tmp_file
            vcf_file, vcf_index = sample_data.get_vcf_file()
            assert vcf_file.exists()
            assert vcf_index.exists()
            assert vcf_file.parent == tmp_file
            assert vcf_index.parent == tmp_file

            count += 1

        # Make sure I tested all 3 files in above loop
        assert 3 == count


def test_iter_sample_data_mixed_masks(sample_dirs):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        vcf_masks = vcf_and_mixed_mask_files()
        file_processor = SerialSampleFilesProcessor(tmp_file)
        data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                         masked_genomic_files_map=vcf_masks[
                                                                             'masks'],
                                                                         sample_files_processor=file_processor)

        processed_files_dict = {}
        for sample_file in data_package.iter_sample_data():
            processed_files_dict[sample_file.sample_name] = sample_file

        assert 3 == len(processed_files_dict)
        assert {'SampleA', 'SampleB', 'SampleC'} == set(processed_files_dict.keys())

        count = 0
        for sample, mask_len, seq_names in [('SampleA', 437, {'reference'}), ('SampleB', 276, {'reference'}),
                                            ('SampleC', 0, set())]:
            sample_data = processed_files_dict[sample]
            assert isinstance(sample_data, NucleotideSampleData)
            sample_data = cast(NucleotideSampleData, sample_data)
            mask = sample_data.get_mask()
            assert mask_len == len(mask)
            assert seq_names == mask.sequence_names()
            mask_file = sample_data.get_mask_file()
            assert mask_file.parent == tmp_file
            vcf_file, vcf_index = sample_data.get_vcf_file()
            assert vcf_file.exists()
            assert vcf_index.exists()
            assert vcf_file.parent == tmp_file
            assert vcf_index.parent == tmp_file

            count += 1

        # Make sure I tested all 3 files in above loop
        assert 3 == count


def do_test_get_features_reader_and_features_table(data_package: NucleotideSampleDataPackage):
    # If data is not preprocessed, make sure exception is raised
    features_reader = data_package.get_features_reader()
    assert features_reader is not None
    with pytest.raises(Exception) as execinfo:
        features_reader.get_features_table()
    assert 'VCF file for sample' in str(execinfo.value)
    assert 'is not preprocessed' in str(execinfo.value)

    # Process data
    data_package_processed = data_package.process_all_data()

    # Now get features reader from new data package and verify I can read feature tables
    features_reader = data_package_processed.get_features_reader()
    features_df = features_reader.get_features_table()
    assert 1170 == len(features_df)
    assert {'SampleA', 'SampleB', 'SampleC'} == set(features_df['SAMPLE'].tolist())


def test_get_features_reader_and_features_table_serial_variants_processor(sample_dirs):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        vcf_masks = vcf_and_mask_files(sample_dirs)
        file_processor = SerialSampleFilesProcessor(tmp_file)
        variants_processor_factory = SerialVcfVariantsTableProcessorFactory.instance()
        data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                         masked_genomic_files_map=vcf_masks[
                                                                             'masks'],
                                                                         variants_processor_factory=variants_processor_factory,
                                                                         sample_files_processor=file_processor)
        do_test_get_features_reader_and_features_table(data_package)


def test_get_features_reader_and_features_table_parallel_variants_processor(sample_dirs):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        vcf_masks = vcf_and_mask_files(sample_dirs)
        file_processor = SerialSampleFilesProcessor(tmp_file)
        variants_processor_factory = MultipleProcessVcfVariantsTableProcessorFactory(ncores=2)
        data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                         masked_genomic_files_map=vcf_masks[
                                                                             'masks'],
                                                                         variants_processor_factory=variants_processor_factory,
                                                                         sample_files_processor=file_processor)
        do_test_get_features_reader_and_features_table(data_package)


def test_persisted_sample_data_file_names(sample_dirs):
    def split_all_ext(file: Path) -> str:
        return file.name.split('.')[0]

    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        vcf_masks = vcf_and_mask_files(sample_dirs)
        file_processor = SerialSampleFilesProcessor(tmp_file)
        data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                         masked_genomic_files_map=vcf_masks[
                                                                             'masks'],
                                                                         sample_files_processor=file_processor)

        processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
        processed_files_dict = processed_data_package.get_sample_data()
        sample_data = cast(NucleotideSampleData, processed_files_dict['SampleA'])
        mask_file = sample_data.get_mask_file()
        vcf_file, vcf_index = sample_data.get_vcf_file()

        assert sample_data.sample_name_persistence == split_all_ext(mask_file)
        assert sample_data.sample_name_persistence == split_all_ext(vcf_file)
        assert sample_data.sample_name_persistence == split_all_ext(vcf_index)
        assert 32 == len(sample_data.sample_name_persistence)
        assert bool(re.match(r'^[0-9a-f]*$', sample_data.sample_name_persistence))


def test_with_serial_sample_files_processor(sample_dirs):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        vcf_masks = vcf_and_mask_files(sample_dirs)
        file_processor = SerialSampleFilesProcessor(tmp_file)
        data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                         masked_genomic_files_map=vcf_masks[
                                                                             'masks'],
                                                                         sample_files_processor=file_processor)

        processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
        processed_files_dict = processed_data_package.get_sample_data()

        assert 3 == len(processed_files_dict)

        sample_data = processed_files_dict['SampleA']
        assert isinstance(sample_data, NucleotideSampleData)
        sample_data = cast(NucleotideSampleData, sample_data)
        mask = sample_data.get_mask()
        assert 437 == len(mask)
        assert {'reference'} == mask.sequence_names()
        assert sample_data.get_mask_file().parent == tmp_file
        vcf_file, vcf_index = sample_data.get_vcf_file()
        assert vcf_file.exists()
        assert vcf_index.exists()
        assert vcf_file.parent == tmp_file
        assert vcf_index.parent == tmp_file

        sample_data = processed_files_dict['SampleB']
        assert isinstance(sample_data, NucleotideSampleData)
        sample_data = cast(NucleotideSampleData, sample_data)
        mask = sample_data.get_mask()
        assert 276 == len(mask)
        assert {'reference'} == mask.sequence_names()
        assert sample_data.get_mask_file().parent == tmp_file
        vcf_file, vcf_index = sample_data.get_vcf_file()
        assert vcf_file.exists()
        assert vcf_index.exists()
        assert vcf_file.parent == tmp_file
        assert vcf_index.parent == tmp_file

        sample_data = processed_files_dict['SampleC']
        assert isinstance(sample_data, NucleotideSampleData)
        sample_data = cast(NucleotideSampleData, sample_data)
        mask = sample_data.get_mask()
        assert 329 == len(mask)
        assert {'reference'} == mask.sequence_names()
        assert sample_data.get_mask_file().parent == tmp_file
        vcf_file, vcf_index = sample_data.get_vcf_file()
        assert vcf_file.exists()
        assert vcf_index.exists()
        assert vcf_file.parent == tmp_file
        assert vcf_index.parent == tmp_file


def test_with_multiprocess_sample_files_processor(sample_dirs):
    with TemporaryDirectory() as tmp_file_str:
        tmp_file = Path(tmp_file_str)
        vcf_masks = vcf_and_mask_files(sample_dirs)
        file_processor = MultipleProcessSampleFilesProcessor(tmp_file, processing_cores=2)
        data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                         masked_genomic_files_map=vcf_masks[
                                                                             'masks'],
                                                                         sample_files_processor=file_processor)

        processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
        processed_files_dict = processed_data_package.get_sample_data()

        assert 3 == len(processed_files_dict)

        sample_data = processed_files_dict['SampleA']
        assert isinstance(sample_data, NucleotideSampleData)
        sample_data = cast(NucleotideSampleData, sample_data)
        mask = sample_data.get_mask()
        assert 437 == len(mask)
        assert {'reference'} == mask.sequence_names()
        assert sample_data.get_mask_file().parent == tmp_file
        vcf_file, vcf_index = sample_data.get_vcf_file()
        assert vcf_file.exists()
        assert vcf_index.exists()
        assert vcf_file.parent == tmp_file
        assert vcf_index.parent == tmp_file

        sample_data = processed_files_dict['SampleB']
        assert isinstance(sample_data, NucleotideSampleData)
        sample_data = cast(NucleotideSampleData, sample_data)
        mask = sample_data.get_mask()
        assert 276 == len(mask)
        assert {'reference'} == mask.sequence_names()
        assert sample_data.get_mask_file().parent == tmp_file
        vcf_file, vcf_index = sample_data.get_vcf_file()
        assert vcf_file.exists()
        assert vcf_index.exists()
        assert vcf_file.parent == tmp_file
        assert vcf_index.parent == tmp_file

        sample_data = processed_files_dict['SampleC']
        assert isinstance(sample_data, NucleotideSampleData)
        sample_data = cast(NucleotideSampleData, sample_data)
        mask = sample_data.get_mask()
        assert 329 == len(mask)
        assert {'reference'} == mask.sequence_names()
        assert sample_data.get_mask_file().parent == tmp_file
        vcf_file, vcf_index = sample_data.get_vcf_file()
        assert vcf_file.exists()
        assert vcf_index.exists()
        assert vcf_file.parent == tmp_file
        assert vcf_index.parent == tmp_file


def test_with_null_sample_files_processor(sample_dirs):
    vcf_masks = vcf_and_mask_files(sample_dirs)
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                     masked_genomic_files_map=vcf_masks[
                                                                         'masks'])

    processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
    processed_files_dict = processed_data_package.get_sample_data()

    assert 3 == len(processed_files_dict)
    sample_data = processed_files_dict['SampleA']
    assert isinstance(sample_data, NucleotideSampleData)
    sample_data = cast(NucleotideSampleData, sample_data)

    with pytest.raises(Exception) as execinfo:
        sample_data.get_mask()
    assert 'Sample mask file is not preprocessed for sample' in str(execinfo.value)

    with pytest.raises(Exception) as execinfo:
        sample_data.get_vcf_file()
    assert 'VCF file for sample' in str(execinfo.value)



import re
from pathlib import Path

import pandas as pd

from genomics_data_index.storage.io.mlst.MLSTFeaturesReader import MLSTFeaturesReader
from genomics_data_index.storage.model import MLST_UNKNOWN_ALLELE


class MLSTTSeemannFeaturesReader(MLSTFeaturesReader):
    """
    A reader for results from the MLST software developed by Torsten Seemann (https://github.com/tseemann/mlst).
    Assumes output has been produced like:

    mlst --nopath *.fasta > mlst.tsv
    """

    def __init__(self, mlst_file: Path):
        super().__init__()

        self._mlst_file = mlst_file

    def _read_features_table(self) -> pd.DataFrame:
        df = pd.read_csv(self._mlst_file, sep='\t', header=None)
        df = df.rename(columns={
            0: 'File',
            1: 'Scheme',
            2: 'Sequence Type',
        })

        df['Sample'] = self._get_sample_from_filename(df['File'])
        df = self._extract_locus_alleles(df)

        df = df[['File', 'Sample', 'Scheme', 'Locus', 'Allele', 'Sequence Type']].sort_values(
            by=['Sample', 'Scheme', 'Locus']).reset_index().drop(columns='index')

        return df

    def _is_valid_allele(self, allele: str) -> bool:
        return allele != MLST_UNKNOWN_ALLELE and bool(re.match(r'^\d+$', allele))

    def _get_sample_from_filename(self, filename_series: pd.Series) -> pd.Series:
        file_sample_name_regex = r'^([^.]*)'
        return filename_series.str.extract(file_sample_name_regex, expand=True)

    def _extract_locus_alleles(self, df: pd.DataFrame) -> pd.DataFrame:
        locus_allele_list = list(set(df.columns) - {'File', 'Sample', 'Scheme', 'Sequence Type'})
        df = df.melt(id_vars=['File', 'Sample', 'Scheme', 'Sequence Type'], value_vars=locus_allele_list)
        df[['Locus', 'Allele']] = df['value'].str.extract(r'^([^\(]*)\(([^\)]*)\)', expand=True)
        return df



from __future__ import annotations

from typing import List, Optional

from genomics_data_index.storage.model.QueryFeature import QueryFeature
from genomics_data_index.storage.model.QueryFeatureHGVS import QueryFeatureHGVS


class QueryFeatureHGVSGN(QueryFeatureHGVS):
    PREFIX = 'hgvs_gn:'

    def __init__(self, sequence_name: str, gene: str, variant: str):
        super().__init__(sequence_name=sequence_name, gene=gene, variant=variant)

    def is_unknown(self) -> bool:
        raise NotImplementedError('Not implemented')

    def to_unknown(self) -> QueryFeature:
        raise NotImplementedError('Not implemented')

    def to_unknown_explode(self) -> List[QueryFeature]:
        raise NotImplementedError('Not implemented')

    @classmethod
    def create(cls, sequence_name: str, gene: Optional[str], variant: Optional[str]) -> QueryFeatureHGVSGN:
        # If variant is given as nucleotide coordinates with respect to the sequence 'n.' then ignore gene id
        if variant is not None and variant.startswith('n.'):
            gene = None

        return QueryFeatureHGVSGN(sequence_name=sequence_name, gene=gene, variant=variant)

    @classmethod
    def create_from_id(cls, hgvs_id: str, prefix: str = None) -> QueryFeatureHGVSGN:
        """
        Creates a new HGVSGN feature which uses the gene name (instead of locus id) for the "gene" component. This
        is nearly identical to an HGVS identifier, except that the [gene] portion in the id
        (hgvs_gn:[sequence]:[gene]:[variant]) now corresponds to the gene name instead of locus identifier. For example,
        hgvs_gn:MN996528.1:S:p.D614G to specify the D614G mutation on the spike (S) protein name instead of the
        corresponding locus identifier.
        :param hgvs_id: The (modified) HGVS identifier.
        :param prefix: The prefix to use for this identifier. Used to override prefix for HGVSGN
        :return: A new HGVS query feature.
        """
        if prefix is None:
            prefix = cls.PREFIX

        sequence_name, gene, variant = QueryFeatureHGVS.split_id(hgvs_id, prefix=prefix)
        return QueryFeatureHGVSGN(sequence_name=sequence_name, gene=gene, variant=variant)



import logging
from typing import Dict, Any, List

import pandas as pd
from pandas.api.types import CategoricalDtype

from genomics_data_index.storage.util import TRACE_LEVEL

logger = logging.getLogger(__name__)


class InvalidSnpEffVcfError(Exception):

    def __init__(self, message):
        super().__init__(message)


class VcfSnpEffAnnotationParser:
    ANNOTATION_COLUMNS = ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                          'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    # Extra annotation columns are those used internally but aren't going to be stored in the database
    EXTRA_ANNOTATION_COLUMNS = ['ANN.cDNA.pos / cDNA.length']

    IMPACT_TYPE = CategoricalDtype(categories=['HIGH', 'MODERATE', 'LOW', 'MODIFIER'], ordered=True)

    # Order of categories derived from list provided in <http://pcingola.github.io/SnpEff/adds/VCFannotationformat_v1.0.pdf>
    # But has been modified a bit
    ANNOTATION_TYPE = CategoricalDtype(categories=[
        "chromosome_number_variation",
        "exon_loss_variant",
        "frameshift_variant",
        "stop_gained",
        "stop_lost",
        "start_lost",
        "splice_acceptor_variant",
        "splice_donor_variant",
        "rare_amino_acid_variant",
        "missense_variant",
        "disruptive_inframe_insertion",
        "conservative_inframe_insertion",
        "disruptive_inframe_deletion",
        "conservative_inframe_deletion",
        "5_prime_UTR_truncation+exon_loss_variant",
        "3_prime_UTR_truncation+exon_loss",
        "splice_branch_variant",
        "splice_region_variant",
        "stop_retained_variant",
        "initiator_codon_variant",
        "synonymous_variant",
        "initiator_codon_variant+non_canonical_start_codon",
        "coding_sequence_variant",
        "5_prime_UTR_variant",
        "3_prime_UTR_variant",
        "5_prime_UTR_premature_start_codon_gain_variant",
        "intragenic_variant",
        "intergenic_region",
        "upstream_gene_variant",
        "downstream_gene_variant",
        "TF_binding_site_variant",
        "regulatory_region_variant",
        "miRNA",
        "custom",
        "sequence_feature",
        "conserved_intron_variant",
        "intron_variant",
        "conserved_intergenic_variant",
        "non_coding_exon_variant",
        "nc_transcript_variant",
        "gene_variant",
        "chromosome",
    ], ordered=True)

    def __init__(self):
        pass

    def parse_annotation_headers(self, vcf_info: Dict[str, Any]) -> List[str]:
        expected_start_str = "Functional annotations: '"

        if 'ANN' in vcf_info:
            ann_info = vcf_info['ANN'].desc
            if not ann_info.startswith(expected_start_str):
                raise InvalidSnpEffVcfError(f"Found 'ANN' in VCF information but description does "
                                            f"not start with [{expected_start_str}]. desc=[{ann_info}]")
            else:
                ann_info = ann_info[len(expected_start_str):]

                if ann_info.endswith("'"):
                    ann_info = ann_info[0:len(ann_info) - 1]

                return [x.strip() for x in ann_info.split('|')]
        else:
            logger.log(TRACE_LEVEL, f"VCF does not contain 'ANN' in vcf_info, keys=[{vcf_info.keys()}]")
            return []

    def _setup_vcf_df_index(self, vcf_df: pd.DataFrame) -> pd.DataFrame:
        vcf_df_with_keys = vcf_df.copy()
        if len(vcf_df_with_keys) == 0:
            vcf_df_with_keys = vcf_df_with_keys.reindex(columns=vcf_df_with_keys.columns.tolist() + ['VARIANT_ID'])
        else:
            vcf_df_with_keys['VARIANT_ID'] = vcf_df_with_keys.apply(
                lambda x: f"{x['CHROM']}:{x['POS']}:{x['REF']}:{x['ALT']}",
                axis='columns')
            vcf_df_with_keys = vcf_df_with_keys.reset_index().rename({'index': 'original_index'}, axis='columns')
            vcf_df_with_keys = vcf_df_with_keys.set_index('VARIANT_ID')

        return vcf_df_with_keys

    def _extract_ann_from_info(self, vcf_df_with_keys: pd.DataFrame, vcf_ann_headers: List[str]) -> pd.DataFrame:
        ann_groups = vcf_df_with_keys['INFO'].map(lambda x: x['ANN']).explode()
        ann_split_fields = ann_groups.map(lambda x: x.split('|') if not pd.isna(x) else [pd.NA] * len(vcf_ann_headers))
        ann_split_fields = ann_split_fields.map(lambda x: dict(zip(vcf_ann_headers, x))).to_frame()
        ann_split_fields = ann_split_fields.rename({'INFO': 'ANN'}, axis='columns').reset_index()

        return ann_split_fields

    def vcf_has_annotations(self, vcf_df: pd.DataFrame) -> bool:
        return len(vcf_df) > 0 and 'INFO' in vcf_df and 'ANN' in vcf_df.iloc[0]['INFO']

    def parse_annotation_entries(self, vcf_ann_headers: List[str], vcf_df: pd.DataFrame) -> pd.DataFrame:
        """
        Given a list of snpeff VCF 'ANN' headers and the VCF datqframe, splits up the snpeff effects into separate rows.
        Returns a new dataframe with all the snpeff 'ANN' effects, one per row. The index is the same as the input 'vcf_df'
        dataframe which means you can merge the returned value to 'vcf_df' afterwards by the indexes.
        :param vcf_ann_headers: A list of headers for the VCF 'ANN' entries (order matters here).
        :param vcf_df: The dataframe containing the VCF information. The 'ANN' information is assumed to exist
                       in the 'INFO' column.
        :return: A dataframe with the snpeff annotation entries, one entry per line. The index is the same as the input
                 vcf_df, which means that you can merge the returned dataframe with 'vcf_df' by the index. If there are
                 no snpeff annotations then this will return an empty dataframe with the appropriate snpeff ANN columns
                 names.
        """
        vcf_df_with_keys = self._setup_vcf_df_index(vcf_df)

        if self.vcf_has_annotations(vcf_df):
            ann_split_fields = self._extract_ann_from_info(vcf_df_with_keys, vcf_ann_headers)

            def insert_key_to_dictionary(x: pd.Series):
                """
                I use this to insert the 'VARIANT_KEY' into the 'ANN' dictionary
                so that when I run pd.json_normalize I can keep everything linked by the same key
                (since json_normalize removes my index in the Series).
                :param x: A series of columns.
                :return: A series of columns with the 'VARIANT_KEY' inserted into the 'ANN' dictionary.
                """
                x['ANN']['VARIANT_ID'] = x['VARIANT_ID']
                return x

            ann_series = ann_split_fields.apply(insert_key_to_dictionary, axis='columns')['ANN']
            ann_df = pd.json_normalize(ann_series).add_prefix('ANN.').rename(
                {'ANN.VARIANT_ID': 'VARIANT_ID'}, axis='columns').set_index('VARIANT_ID')

            # Copy index to column 'VARIANT_ID' so it remains after merging
            ann_df['VARIANT_ID'] = ann_df.index

            # Now join ann_df back to vcf_df_with_keys to recover the original index
            vcf_df_with_keys = vcf_df_with_keys.merge(ann_df, left_index=True, right_index=True).set_index(
                'original_index')

            vcf_df_with_keys = vcf_df_with_keys[
                self.ANNOTATION_COLUMNS + self.EXTRA_ANNOTATION_COLUMNS + ['VARIANT_ID']]

            # Fill NA with '' and convert '' back to NA because a straight conversion of '' to NA
            # fails in cases where a column has a mixture of '' and NA values.
            vcf_df_with_keys = vcf_df_with_keys.fillna('').replace('', pd.NA)
        else:
            logger.log(TRACE_LEVEL, 'vcf_df has no snpeff annotations, will set all annotations as NA')
            vcf_df_with_keys = vcf_df_with_keys.reset_index().reindex(
                columns=self.ANNOTATION_COLUMNS + self.EXTRA_ANNOTATION_COLUMNS + ['original_index', 'VARIANT_ID']) \
                .set_index('original_index')

        return vcf_df_with_keys

    def _split_annotations_by_cDNA(self, vcf_df: pd.DataFrame) -> pd.DataFrame:
        vcf_df_nona = vcf_df[~vcf_df['ANN.cDNA.pos / cDNA.length'].isna()]

        if len(vcf_df_nona) > 0:
            vcf_cdna_split = vcf_df_nona['ANN.cDNA.pos / cDNA.length'].str.split('/', expand=True).rename(
                {0: 'ANN.cDNA.pos', 1: 'ANN.cDNA.length'}, axis='columns')
            vcf_cdna_split['ANN.cDNA.pos'] = pd.to_numeric(vcf_cdna_split['ANN.cDNA.pos'])
            vcf_cdna_split['ANN.cDNA.length'] = pd.to_numeric(vcf_cdna_split['ANN.cDNA.length'])

            vcf_df = vcf_df.merge(vcf_cdna_split, how='left', left_index=True, right_index=True)
        else:
            vcf_df = vcf_df.copy()

            vcf_df['ANN.cDNA.pos'] = pd.NA
            vcf_df['ANN.cDNA.length'] = pd.NA

        return vcf_df

    def select_variant_annotations(self, vcf_df: pd.DataFrame) -> pd.DataFrame:
        """
        Given a dataframe containing variants from a VCF file, select the appropriate annotations.
        :param vcf_df: The dataframe of variants from a VCF file.
        :return: A new dataframe containing only rows with the selected annotations.
        """
        non_annotation_columns = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID']
        expected_columns_set = set(non_annotation_columns + self.ANNOTATION_COLUMNS + self.EXTRA_ANNOTATION_COLUMNS)
        actual_columns_set = set(vcf_df.columns.tolist())
        if expected_columns_set != actual_columns_set:
            raise Exception(f'Invalid columns for passed vcf_df.'
                            f' Unique expected={expected_columns_set - actual_columns_set}.'
                            f' Unique actual = {actual_columns_set - expected_columns_set}')

        # Use .copy() to get rid of warnings like "A value is trying to be set on a copy of a slice from a DataFrame"
        vcf_df_with_annotations = vcf_df[~vcf_df['ANN.Allele'].isna()].copy()

        # Set ordered types
        vcf_df_with_annotations['ANN.Annotation_Impact'] = vcf_df_with_annotations['ANN.Annotation_Impact'].astype(
            self.IMPACT_TYPE
        )
        vcf_df_with_annotations['ANN.Annotation'] = vcf_df_with_annotations['ANN.Annotation'].astype(
            self.ANNOTATION_TYPE
        )

        # Remove annotations from being considered
        vcf_df_with_annotations = vcf_df_with_annotations[
            vcf_df_with_annotations['ANN.Allele'] == vcf_df_with_annotations['ALT']]

        # Add columns for ANN.cDNA.pos and ANN.cDNA.length
        vcf_df_with_annotations = self._split_annotations_by_cDNA(vcf_df_with_annotations)

        # Order and select first entry (use nth() instead of first() so that NA values are handled properly)
        # I order by ANN.cDNA.length in descending order so that I pick the largest cDNA.length first
        # (this solves issues with picking the correct annotation for ORF1ab and SARS-CoV-2
        # but should work for other organisms as well).
        vcf_df_with_annotations = vcf_df_with_annotations.sort_values(
            ['ANN.Annotation_Impact', 'ANN.Annotation', 'ANN.cDNA.length'],
            ascending=[True, True, False]).groupby('VARIANT_ID').nth(0).reset_index()

        # Merge back with original dataframe of variants to make sure I include those without annotations
        all_vcf_entries_grouped = vcf_df[non_annotation_columns].groupby('VARIANT_ID').first()
        return_vcf_df = all_vcf_entries_grouped.merge(vcf_df_with_annotations, how='left', left_on='VARIANT_ID',
                                                      right_on='VARIANT_ID',
                                                      suffixes=(None, '_with_annotations'))

        return_vcf_df = return_vcf_df[non_annotation_columns + self.ANNOTATION_COLUMNS].sort_values(
            'POS')

        return return_vcf_df



from typing import cast, List, Union

from ete3 import ClusterTree, Tree

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.TreeSamplesQuery import TreeSamplesQuery
from genomics_data_index.api.query.impl.WrappedSamplesQuery import WrappedSamplesQuery
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet


class KmerTreeSamplesQuery(TreeSamplesQuery):
    """
    A KmerTreeSamplesQuery defines a query with a joined tree derived from kmers.
    """

    def __init__(self, connection: DataIndexConnection, wrapped_query: SamplesQuery, tree: ClusterTree):
        """
        Builds a new KmerTreeSamplesQuery from the given information. In most normal operations SamplesQuery objects
        are not created directly but are instead created from an :py:class:`genomics_data_index.api.GenomicsDataIndex`
        object or from operations applied to a SamplesQuery (e.g., build_tree() or join_tree()).

        :param connection: A connection to a database containing samples.
        :param wrapped_query: The SamplesQuery to wrap around and join to the passed tree.
        :param tree: The tree to join to this query.
        :return: A new KmerTreeSamplesQuery object.
        """
        super().__init__(connection=connection, wrapped_query=wrapped_query, tree=tree)

    def _wrap_create(self, wrapped_query: SamplesQuery, universe_set: SampleSet = None) -> WrappedSamplesQuery:
        tree = self.tree
        if isinstance(tree, ClusterTree):
            tree = cast(ClusterTree, self.tree)
        else:
            raise Exception(f'Incorrect type of tree [{tree}]. Expected [{ClusterTree.__class__}], got [{type(tree)}].')

        return KmerTreeSamplesQuery(connection=self._query_connection,
                                    wrapped_query=wrapped_query,
                                    tree=tree)

    def _create_from_tree_internal(self, tree: Tree) -> SamplesQuery:
        if isinstance(tree, ClusterTree):
            tree = cast(ClusterTree, tree)
        else:
            raise Exception(f'Incorrect type of tree [{tree}]. Expected [{ClusterTree.__class__}], got [{type(tree)}].')

        return KmerTreeSamplesQuery(connection=self._query_connection,
                                    wrapped_query=self._wrapped_query,
                                    tree=tree)

    # Override primarily to specify a default value for 'units'
    def _within_distance(self, data: Union[str, List[str], SamplesQuery, SampleSet], distance: float,
                         units: str = 'kmer_jaccard', **kwargs) -> SamplesQuery:
        return super()._within_distance(data=data, distance=distance,
                                        units=units, **kwargs)

    def _within_distance_internal(self, sample_names: Union[str, List[str]], distance: float,
                                  units: str) -> SamplesQuery:
        raise NotImplementedError(f'Not implemented for {KmerTreeSamplesQuery.__class__}')

    def _can_handle_distance_units(self, units: str) -> bool:
        return False

    def _distance_units(self) -> List[str]:
        return self._wrapped_query._distance_units()



import os
from pathlib import Path

assemblies_dir = Path(__file__).parent / 'data' / 'assemblies'
reads_dir = Path(__file__).parent / 'data' / 'reads'
samples_dir = assemblies_dir / 'samples'
expected_mutations_dir = assemblies_dir / 'expected'
assemblies_reference = assemblies_dir / 'genome.fasta'
samples = [os.path.splitext(s)[0] for s in os.listdir(samples_dir)]
assemblies_samples = {s: samples_dir / f'{s}.fasta' for s in samples}
expected_mutations = {s: expected_mutations_dir / f'{s}-mutations.txt' for s in samples}
expected_beds = {
    'SampleD': expected_mutations_dir / 'SampleD.bed.gz',
    'SampleE': expected_mutations_dir / 'SampleE.bed.gz',
    'SampleF': expected_mutations_dir / 'SampleF.bed.gz',
    'SampleH': expected_mutations_dir / 'SampleH.bed.gz',
}

empty_bed_file = expected_mutations_dir / 'empty.bed.gz'

snpeff_input_sampleA = Path(__file__).parent / '..' / 'data' / 'snpeff-database' / 'input' / 'SampleA.fasta.gz'
snpeff_reference_genome = Path(__file__).parent / '..' / 'data' / 'snpeff-database' / 'NC_011083-5000.gbk.gz'

snpeff_reads_paired = [reads_dir / 'SampleA-snpeff_1.fastq.gz', reads_dir / 'SampleA-snpeff_2.fastq.gz']
snpeff_reads_single = [reads_dir / 'SampleA-single-snpeff.fastq.gz']



from typing import Union

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.cluster.ClusterScoreMethod import ClusterScoreMethod
from genomics_data_index.storage.SampleSet import SampleSet


class ClusterScoreMRCAJaccard(ClusterScoreMethod):
    """
    A method which assigns cluster scores based on all the samples under the most recent common ancestor (MRCA) of the
    passed set of samples.
    """

    def __init__(self):
        super().__init__()

    def score(self, data: Union[SamplesQuery, SampleSet], universe_samples: SamplesQuery) -> float:
        if isinstance(data, SamplesQuery):
            data_set = data.sample_set
        elif isinstance(data, SampleSet):
            data_set = data
        else:
            raise Exception(f'Invalid type for data={data}. Got type {type(data)}. Expected {SamplesQuery.__name__}'
                            f' or {SampleSet.__name__}')

        mrca_samples = universe_samples.isin(data, kind='mrca')
        return data_set.jaccard_index(mrca_samples.sample_set)



import copy
import logging
from typing import Union, Iterable, Dict, Any

from ete3 import Tree, TreeStyle, TextFace

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.viewer.samples_visuals.AbstractAnnotationTreeSamplesVisual import \
    AbstractAnnotationTreeSamplesVisual

logger = logging.getLogger(__name__)


class AnnotateTreeSamplesVisual(AbstractAnnotationTreeSamplesVisual):
    """
    Creates a visual style which adds boxes which are colored based on the presence/absence of samples in the set.
    """

    def __init__(self,
                 samples: Union[SamplesQuery, Iterable[str]],
                 label: Union[str, Dict[str, Any]],
                 label_unknown: Union[str, Dict[str, Any]],
                 annotate_show_box_label: bool,
                 annotate_box_label_color: str,
                 annotate_box_label_unknown_color: str,
                 annotate_label_fontsize: int,
                 legend_label: str,
                 box_width: int, box_height: int,
                 color_present: str, color_absent: str, color_unknown: str,
                 legend_nodesize: int,
                 legend_fontsize: int,
                 annotate_column: int,
                 annotate_kind: str,
                 annotate_border_color: str,
                 annotate_opacity_present: float,
                 annotate_opacity_absent: float,
                 annotate_opacity_unknown: float,
                 border_width: int,
                 margin: int,
                 include_unknown: bool,
                 legend_columns: Dict[str, int]):
        """
        Creates a new AnnotateTreeSamplesVisual with the given information. This is not intended to be created directly
        but instead is created via methods in the :py:class:`genomics_data_index.api.viewer.TreeStyler` class.
        :param samples: The set of samples this visual style applies to.
        :param label: A label for the this set of samples.
        :param label_unknown: A label for the unknown set of samples.
        :param annotate_show_box_label: Whether or not the label should be printed in each annotate box.
        :param annotate_box_label_color: The color of the annotate box label if it is to be displayed.
        :param annotate_box_label_unknown_color: The color of the annotate box label (for unknowns) if it is to be displayed.
        :param annotate_label_fontsize: The font size of the annotate box label if it is to be displayed.
        :param legend_label: The legend label.
        :param box_width: The width of the annotate box.
        :param box_height: The height of the annotate box.
        :param color_present: The color of the samples present in this set.
        :param color_absent: The color of the samples absent in this set.
        :param color_unknown: The color of the samples which are unknown if present/absent in this set.
        :param legend_nodesize: The node size for the legend.
        :param legend_fontsize: The legend font size.
        :param annotate_column: Which column in the annotate table this annotate visual belongs to (starting from column 0).
        :param annotate_kind: The type of annotation to display for present samples ['circle', 'rectangle'].
        :param annotate_border_color: The border color of teh annotate box.
        :param annotate_opacity_present: The opacity of the present samples.
        :param annotate_opacity_absent: The opacity of the absent samples.
        :param annotate_opacity_unknown: The opacity of the unknown samples.
        :param border_width: The width of the border.
        :param margin: The width of the margin in each annotate box.
        :param include_unknown: Whether or not unknowns should be included/represented.
        :param legend_columns: A dict mapping legend item types to column numbers.
        :return: A new AnnotateTreeSamplesVisual.
        """
        super().__init__(samples=samples,
                         legend_nodesize=legend_nodesize,
                         legend_fontsize=legend_fontsize,
                         legend_columns=legend_columns,
                         annotate_kind=annotate_kind,
                         annotate_border_width=border_width,
                         annotate_margin=margin)
        self._annotate_show_box_label = annotate_show_box_label
        self._annotate_label_fontsize = annotate_label_fontsize
        self._label = label
        self._legend_label = legend_label
        self._box_width = box_width
        self._box_height = box_height
        self._color_present = color_present
        self._color_absent = color_absent
        self._color_unknown = color_unknown
        self._annotate_column = annotate_column

        self._annotate_border_color = annotate_border_color
        self._annotate_opacity_present = annotate_opacity_present
        self._annotate_opacity_absent = annotate_opacity_absent
        self._annotate_opacity_unknown = annotate_opacity_unknown
        self._include_unknown = include_unknown

        if self._annotate_show_box_label:
            self._box_label_present = copy.deepcopy(label)
            self._box_label_unknown = copy.deepcopy(label_unknown)
            if self._box_label_present is not None and annotate_box_label_color is not None:
                self._box_label_present['color'] = annotate_box_label_color
            if self._box_label_unknown is not None and annotate_box_label_unknown_color is not None:
                self._box_label_unknown['color'] = annotate_box_label_unknown_color
        else:
            self._box_label_present = None
            self._box_label_unknown = None

    def apply_visual(self, tree: Tree, tree_style: TreeStyle) -> None:
        if self._label is not None:
            text = self._label.get('text', None)
            fsize = self._label.get('fontsize', self._annotate_label_fontsize)
            ftype = self._label.get('font', 'Verdana')
            color = self._label.get('color', 'black')
            tf = TextFace(text, fsize=fsize, ftype=ftype, fgcolor=color)
            tf.margin_bottom = 10
            tf.margin_left = 10
            tf.margin_right = 10
            tf.margin_top = 10
            tf.hz_align = 1
            tree_style.aligned_header.add_face(tf, self._annotate_column)
            tree_style.aligned_foot.add_face(tf, self._annotate_column)

        # Annotate nodes
        unknown_samples_count = 0
        for leaf in tree.iter_leaves():
            if leaf.name in self.present_sample_names:
                annotate_face = self._build_annotate_face(width=self._box_width, height=self._box_height,
                                                          border_color=self._annotate_border_color,
                                                          bgcolor=self._color_present,
                                                          opacity=self._annotate_opacity_present,
                                                          label=self._box_label_present)
            elif self._include_unknown and (leaf.name in self.unknown_sample_names):
                annotate_face = self._build_annotate_face(width=self._box_width, height=self._box_height,
                                                          border_color=self._annotate_border_color,
                                                          bgcolor=self._color_unknown,
                                                          opacity=self._annotate_opacity_unknown,
                                                          label=self._box_label_unknown)
                unknown_samples_count = unknown_samples_count + 1
            else:
                annotate_face = self._build_annotate_face(width=self._box_width, height=self._box_height,
                                                          border_color=self._annotate_border_color,
                                                          bgcolor=self._color_absent,
                                                          opacity=self._annotate_opacity_absent,
                                                          label=None)

            leaf.add_face(annotate_face, column=self._annotate_column, position='aligned')

        has_unknowns = unknown_samples_count > 0
        self._add_legend_entry(self._legend_label, legend_color=self._color_present,
                               unknown_color=self._color_unknown, include_unknown=self._include_unknown,
                               has_unknowns=has_unknowns,
                               kind=self._annotate_kind, tree_style=tree_style)



import abc
from typing import Union, List

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.kind.IsaKind import IsaKind
from genomics_data_index.storage.SampleSet import SampleSet


class DelegateIsaKind(IsaKind, abc.ABC):

    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def isa(self, data: Union[str, List[str], SamplesQuery, SampleSet], query: SamplesQuery) -> SamplesQuery:
        pass



from typing import cast

import pandas as pd
import pytest

from genomics_data_index.storage.io.mlst.MLSTFeaturesReader import MLSTFeaturesReader
from genomics_data_index.storage.io.mlst.MLSTSampleData import MLSTSampleData
from genomics_data_index.storage.io.mlst.MLSTSampleDataPackage import MLSTSampleDataPackage
from genomics_data_index.storage.io.mlst.MLSTTSeemannFeaturesReader import MLSTTSeemannFeaturesReader
from genomics_data_index.test.integration import basic_mlst_file


@pytest.fixture
def mlst_reader() -> MLSTTSeemannFeaturesReader:
    return MLSTTSeemannFeaturesReader(mlst_file=basic_mlst_file)


@pytest.fixture
def sample_data_package(mlst_reader: MLSTFeaturesReader) -> MLSTSampleDataPackage:
    return MLSTSampleDataPackage(mlst_reader=mlst_reader)


def test_get_sample_names(sample_data_package: MLSTSampleDataPackage):
    assert {'2014C-3598', '2014C-3599', '2014D-0067', '2014D-0068',
            'CFSAN002349', 'CFSAN023463'} == sample_data_package.sample_names()


def test_iter_samples(sample_data_package: MLSTSampleDataPackage):
    sample_data_map = {}
    for sample_data in sample_data_package.iter_sample_data():
        sample_data_map[sample_data.sample_name] = cast(MLSTSampleData, sample_data)

    assert 6 == len(sample_data_map)
    assert {'2014C-3598', '2014C-3599', '2014D-0067', '2014D-0068',
            'CFSAN002349', 'CFSAN023463'} == set(sample_data_map.keys())

    allele_data = pd.Series({'adk': '100', 'fumC': '23', 'gyrB': '68', 'icd': '45',
                             'mdh': '1', 'purA': '35', 'recA': '7'}, name='Allele')
    allele_data.index.name = 'Locus'
    assert allele_data.equals(sample_data_map['2014C-3598'].get_alleles())



from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


class SamplesProgressLogger:
    """
    Used encapsulate progress monitoring as I'm loading in data.
    """
    total_batches = None
    current_batch = 0
    total_stages = 2

    def __init__(self, stage_name: str, stage_number: int, total_samples: int):
        self._total_samples = total_samples
        self._stage_name = stage_name
        self._stage_number = stage_number

    def get_update_every_nth_sample(self, percent_to_update: float) -> int:
        if percent_to_update < 0 or percent_to_update > 100:
            raise Exception(f'percent_to_update={percent_to_update} must be between 0 and 100')

        sample_to_update = int(self._total_samples * (percent_to_update / 100))
        return max(1, sample_to_update)

    def is_batch_mode(self) -> bool:
        return self.total_batches is not None

    def update_progress(self, sample_number: int) -> None:
        if self.is_batch_mode():
            prefix = f'Sample batch {self.current_batch}/{self.total_batches}: '
        else:
            prefix = ''
        logger.info(f'{prefix}Stage {self._stage_number}/{self.total_stages} ({self._stage_name}): '
                    f'Processed {sample_number / self._total_samples * 100:0.0f}% '
                    f'({sample_number}/{self._total_samples}) samples')

    @classmethod
    def set_total_batches(cls, total_batches: int) -> None:
        cls.total_batches = total_batches

    @classmethod
    def set_current_batch(cls, current_batch: int) -> None:
        cls.current_batch = current_batch

    @classmethod
    def unset_batch_mode(cls) -> None:
        cls.total_batches = None



from __future__ import annotations

from genomics_data_index.storage.model.QueryFeature import QueryFeature
from genomics_data_index.storage.model.QueryFeatureHGVS import QueryFeatureHGVS
from genomics_data_index.storage.model.QueryFeatureHGVSGN import QueryFeatureHGVSGN
from genomics_data_index.storage.model.QueryFeatureMLST import QueryFeatureMLST
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI


class QueryFeatureFactory:
    factory_instance = None

    def __init__(self):
        # Maps the prefix string (e.g. 'hgvs:') to a function used to create the specific QueryFeature from the ID
        self._registered_feature_creators = {
            QueryFeatureHGVS.PREFIX: lambda i: QueryFeatureHGVS.create_from_id(i),
            QueryFeatureHGVSGN.PREFIX: lambda i: QueryFeatureHGVSGN.create_from_id(i),
            QueryFeatureMLST.PREFIX: lambda i: QueryFeatureMLST.create_from_id(i),
        }

    def create_feature(self, id: str) -> QueryFeature:
        if id is None:
            raise Exception('id is None')

        for prefix in self._registered_feature_creators:
            if id.startswith(prefix):
                return self._registered_feature_creators[prefix](id)

        # By default assume feature is an SPDI feature
        try:
            return QueryFeatureMutationSPDI(id)
        except Exception as e:
            raise Exception(f'Unknown feature type for feature=[{id}]', e)

    @classmethod
    def instance(cls) -> QueryFeatureFactory:
        if cls.factory_instance is None:
            cls.factory_instance = QueryFeatureFactory()
        return cls.factory_instance



from ete3 import Tree

from genomics_data_index.api.query.impl.TreeBuilderReferenceMutations import TreeBuilderReferenceMutations
from genomics_data_index.test.integration import tree_file

expected_tree = Tree(str(tree_file))


def test_build_tree(loaded_database_connection):
    tree_builder = TreeBuilderReferenceMutations(database_connection=loaded_database_connection,
                                                 reference_name='genome')

    actual_tree, alignment_length, tree_set = tree_builder.build(samples_set=['SampleA', 'SampleB', 'SampleC'],
                                                                 method='iqtree',
                                                                 align_type='full',
                                                                 include_reference=True,
                                                                 extra_params='--seed 42 -m GTR'
                                                                 )

    assert 5180 == alignment_length

    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(actual_tree.get_leaf_names())

    tree_comparison = expected_tree.compare(actual_tree, unrooted=True)
    assert tree_comparison['rf'] == 0

    actual_distance = 5180 * actual_tree.get_distance('SampleA', 'genome')
    expected_distance = 26
    assert abs(expected_distance - actual_distance) < 6


def test_build_tree_core(loaded_database_connection):
    tree_builder = TreeBuilderReferenceMutations(database_connection=loaded_database_connection,
                                                 reference_name='genome')

    actual_tree, alignment_length, tree_set = tree_builder.build(samples_set=['SampleA', 'SampleB', 'SampleC'],
                                                                 method='iqtree',
                                                                 align_type='core',
                                                                 include_variants=['SNP'],
                                                                 include_reference=True,
                                                                 extra_params='--seed 42 -m GTR+ASC'
                                                                 )

    # See test_CoreAlignmentService.py for where my test alignment is coming from and how I get the length
    assert 58 == alignment_length

    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(actual_tree.get_leaf_names())

    tree_comparison = expected_tree.compare(actual_tree, unrooted=True)
    assert tree_comparison['rf'] == 0

    actual_distance = 5180 * actual_tree.get_distance('SampleA', 'genome')
    expected_distance = 26
    assert abs(expected_distance - actual_distance) < 6


def test_build_tree_exclude_reference(loaded_database_connection):
    tree_builder = TreeBuilderReferenceMutations(database_connection=loaded_database_connection,
                                                 reference_name='genome')

    actual_tree, alignment_length, tree_set = tree_builder.build(samples_set=['SampleA', 'SampleB', 'SampleC'],
                                                                 method='iqtree',
                                                                 align_type='full',
                                                                 include_reference=False,
                                                                 extra_params='--seed 42 -m GTR'
                                                                 )

    assert {'SampleA', 'SampleB', 'SampleC'} == set(actual_tree.get_leaf_names())



import tempfile
import warnings
from pathlib import Path

import pytest

warnings.filterwarnings("ignore", category=DeprecationWarning)

from genomics_data_index.test.integration import sample_dirs, reference_file, regular_vcf_dir, data_dir
from genomics_data_index.test.integration import sample_dirs_AB, sample_dirs_C
from genomics_data_index.test.integration import mlst_file_single_scheme, basic_mlst_file, mlst_file_unknown
from genomics_data_index.test.integration import mlst_file_single_scheme2, mlst_file_single_scheme3
from genomics_data_index.test.integration import sourmash_signatures, snpeff_sample_vcfs, reference_file_snpeff
from genomics_data_index.test.integration import snpeff_sample_vcfs_fake_dup
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service.ReferenceService import ReferenceService
from genomics_data_index.storage.service.VariationService import VariationService
from genomics_data_index.storage.service.CoreAlignmentService import CoreAlignmentService
from genomics_data_index.storage.io.mlst.MLSTFeaturesReader import MLSTFeaturesReader
from genomics_data_index.storage.io.mlst.MLSTTSeemannFeaturesReader import MLSTTSeemannFeaturesReader
from genomics_data_index.storage.service.SampleService import SampleService
from genomics_data_index.storage.service.TreeService import TreeService
from genomics_data_index.storage.service.KmerService import KmerService
from genomics_data_index.configuration.connector.FilesystemStorage import FilesystemStorage
from genomics_data_index.storage.service.MLSTService import MLSTService
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.mlst.MLSTSampleDataPackage import MLSTSampleDataPackage
from genomics_data_index.storage.model.db.DatabasePathTranslator import DatabasePathTranslator
from genomics_data_index.storage.io.mutation.variants_processor.MultipleProcessVcfVariantsTableProcessor import \
    MultipleProcessVcfVariantsTableProcessorFactory


@pytest.fixture
def filesystem_storage() -> FilesystemStorage:
    return FilesystemStorage(Path(tempfile.mkdtemp(prefix='index-test')))


@pytest.fixture
def database(filesystem_storage) -> DatabaseConnection:
    dpt = DatabasePathTranslator(filesystem_storage.root_dir)
    return DatabaseConnection('sqlite:///:memory:', dpt)


@pytest.fixture
def reference_service(database, filesystem_storage) -> ReferenceService:
    reference_service = ReferenceService(database, filesystem_storage.reference_dir)
    return reference_service


@pytest.fixture
def snippy_nucleotide_data_package() -> NucleotideSampleDataPackage:
    tmp_dir = Path(tempfile.mkdtemp())
    return NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                          sample_files_processor=SerialSampleFilesProcessor(tmp_dir))


@pytest.fixture
def snippy_nucleotide_data_package_no_index_missing() -> NucleotideSampleDataPackage:
    tmp_dir = Path(tempfile.mkdtemp())
    return NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                          sample_files_processor=SerialSampleFilesProcessor(tmp_dir),
                                                          index_unknown_missing=False)


@pytest.fixture
def snippy_nucleotide_data_package_AB() -> NucleotideSampleDataPackage:
    tmp_dir = Path(tempfile.mkdtemp())
    return NucleotideSampleDataPackage.create_from_snippy(sample_dirs_AB,
                                                          sample_files_processor=SerialSampleFilesProcessor(tmp_dir))


@pytest.fixture
def snippy_nucleotide_data_package_C() -> NucleotideSampleDataPackage:
    tmp_dir = Path(tempfile.mkdtemp())
    return NucleotideSampleDataPackage.create_from_snippy(sample_dirs_C,
                                                          sample_files_processor=SerialSampleFilesProcessor(tmp_dir))


@pytest.fixture
def regular_nucleotide_data_package() -> NucleotideSampleDataPackage:
    vcf_files = {
        'SampleA': Path(regular_vcf_dir, 'SampleA.vcf.gz'),
        'SampleB': Path(regular_vcf_dir, 'SampleB.vcf.gz'),
        'SampleC': Path(regular_vcf_dir, 'SampleC.vcf.gz'),
    }

    mask_files = {
        'SampleA': Path(data_dir, 'SampleA', 'snps.aligned.minus-vcf.bed.gz'),
        'SampleB': Path(data_dir, 'SampleB', 'snps.aligned.minus-vcf.bed.gz'),
        'SampleC': Path(data_dir, 'SampleC', 'snps.aligned.minus-vcf.bed.gz'),
    }

    tmp_dir = Path(tempfile.mkdtemp())
    return NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_files,
                                                             masked_genomic_files_map=mask_files,
                                                             sample_files_processor=SerialSampleFilesProcessor(
                                                                 tmp_dir))


@pytest.fixture
def snpeff_nucleotide_data_package() -> NucleotideSampleDataPackage:
    tmp_dir = Path(tempfile.mkdtemp())
    return NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snpeff_sample_vcfs,
                                                             masked_genomic_files_map=None,
                                                             sample_files_processor=SerialSampleFilesProcessor(
                                                                 tmp_dir))


@pytest.fixture
def snpeff_nucleotide_data_package_fake_duplicate_gene() -> NucleotideSampleDataPackage:
    tmp_dir = Path(tempfile.mkdtemp())
    return NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snpeff_sample_vcfs_fake_dup,
                                                             masked_genomic_files_map=None,
                                                             sample_files_processor=SerialSampleFilesProcessor(
                                                                 tmp_dir))


@pytest.fixture
def snpeff_nucleotide_data_package_parallel_variants() -> NucleotideSampleDataPackage:
    tmp_dir = Path(tempfile.mkdtemp())
    variants_processor_factory = MultipleProcessVcfVariantsTableProcessorFactory(ncores=2)
    return NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snpeff_sample_vcfs,
                                                             masked_genomic_files_map=None,
                                                             sample_files_processor=SerialSampleFilesProcessor(
                                                                 tmp_dir),
                                                             variants_processor_factory=variants_processor_factory)


@pytest.fixture
def reference_service_with_data(reference_service) -> ReferenceService:
    reference_service.add_reference_genome(reference_file)
    return reference_service


@pytest.fixture
def reference_service_with_snpeff_data(reference_service) -> ReferenceService:
    reference_service.add_reference_genome(reference_file_snpeff)
    return reference_service


@pytest.fixture
def sample_service(database):
    return SampleService(database, sql_select_limit=500)


@pytest.fixture
def sample_service_select1(database):
    return SampleService(database, sql_select_limit=1)


@pytest.fixture
def sample_service_select2(database):
    return SampleService(database, sql_select_limit=2)


@pytest.fixture
def sample_service_snpeff_annotations(database, reference_service_with_snpeff_data,
                                      snpeff_nucleotide_data_package, sample_service,
                                      filesystem_storage) -> SampleService:
    var_service = VariationService(database_connection=database,
                                   reference_service=reference_service_with_snpeff_data,
                                   sample_service=sample_service,
                                   variation_dir=filesystem_storage.variation_dir,
                                   index_unknown_missing=False,
                                   sql_select_limit=500)
    var_service.insert(feature_scope_name='NC_011083',
                       data_package=snpeff_nucleotide_data_package)
    return sample_service


@pytest.fixture
def sample_service_snpeff_annotations_fake_duplicate_gene(database, reference_service_with_snpeff_data,
                                                          snpeff_nucleotide_data_package_fake_duplicate_gene,
                                                          sample_service,
                                                          filesystem_storage) -> SampleService:
    var_service = VariationService(database_connection=database,
                                   reference_service=reference_service_with_snpeff_data,
                                   sample_service=sample_service,
                                   variation_dir=filesystem_storage.variation_dir,
                                   index_unknown_missing=False,
                                   sql_select_limit=500)
    var_service.insert(feature_scope_name='NC_011083',
                       data_package=snpeff_nucleotide_data_package_fake_duplicate_gene)
    return sample_service


@pytest.fixture
def sample_service_snpeff_annotations_parallel_variants(database, reference_service_with_snpeff_data,
                                                        snpeff_nucleotide_data_package_parallel_variants,
                                                        sample_service,
                                                        filesystem_storage) -> SampleService:
    var_service = VariationService(database_connection=database,
                                   reference_service=reference_service_with_snpeff_data,
                                   sample_service=sample_service,
                                   variation_dir=filesystem_storage.variation_dir,
                                   index_unknown_missing=False,
                                   sql_select_limit=500)
    var_service.insert(feature_scope_name='NC_011083',
                       data_package=snpeff_nucleotide_data_package_parallel_variants)
    return sample_service


@pytest.fixture
def variation_service(database, reference_service_with_data,
                      snippy_nucleotide_data_package, sample_service, filesystem_storage) -> VariationService:
    var_service = VariationService(database_connection=database,
                                   reference_service=reference_service_with_data,
                                   sample_service=sample_service,
                                   variation_dir=filesystem_storage.variation_dir,
                                   index_unknown_missing=False,
                                   sql_select_limit=500)
    var_service.insert(feature_scope_name='genome',
                       data_package=snippy_nucleotide_data_package)
    return var_service


@pytest.fixture
def variation_service_index_unknowns(database, reference_service_with_data,
                                     snippy_nucleotide_data_package, sample_service,
                                     filesystem_storage) -> VariationService:
    var_service = VariationService(database_connection=database,
                                   reference_service=reference_service_with_data,
                                   sample_service=sample_service,
                                   variation_dir=filesystem_storage.variation_dir,
                                   index_unknown_missing=True,
                                   sql_select_limit=500)
    var_service.insert(feature_scope_name='genome',
                       data_package=snippy_nucleotide_data_package)
    return var_service


@pytest.fixture
def variation_service_non_snippy_vcfs(database, reference_service_with_data,
                                      regular_nucleotide_data_package, sample_service,
                                      filesystem_storage) -> VariationService:
    var_service = VariationService(database_connection=database,
                                   reference_service=reference_service_with_data,
                                   sample_service=sample_service,
                                   variation_dir=filesystem_storage.variation_dir,
                                   index_unknown_missing=False,
                                   sql_select_limit=500)
    var_service.insert(feature_scope_name='genome',
                       data_package=regular_nucleotide_data_package)
    return var_service


@pytest.fixture
def core_alignment_service(database, reference_service_with_data, variation_service,
                           sample_service) -> CoreAlignmentService:
    return CoreAlignmentService(database=database,
                                reference_service=reference_service_with_data,
                                variation_service=variation_service,
                                sample_service=sample_service,
                                )


@pytest.fixture
def core_alignment_service_non_snippy_vcfs(database, reference_service_with_data, variation_service_non_snippy_vcfs,
                                           sample_service) -> CoreAlignmentService:
    return CoreAlignmentService(database=database,
                                reference_service=reference_service_with_data,
                                variation_service=variation_service_non_snippy_vcfs,
                                sample_service=sample_service,
                                )


@pytest.fixture
def tree_service(database, reference_service_with_data, core_alignment_service) -> TreeService:
    return TreeService(database, reference_service_with_data, core_alignment_service)


@pytest.fixture
def tree_service_with_tree_stored(database, reference_service_with_data,
                                  core_alignment_service, variation_service) -> TreeService:
    tree_service = TreeService(database, reference_service_with_data, core_alignment_service)
    tree_service.rebuild_tree('genome',
                              tree_build_type='iqtree',
                              extra_params='-m MFP+ASC --seed 42')

    return tree_service


@pytest.fixture
def kmer_service_with_data(database, sample_service, filesystem_storage) -> KmerService:
    kmer_service = KmerService(database_connection=database,
                               sample_service=sample_service,
                               features_dir=filesystem_storage.kmer_dir)
    for sample_name in sourmash_signatures:
        kmer_service.insert_kmer_index(sample_name=sample_name,
                                       kmer_index_path=sourmash_signatures[sample_name])

    return kmer_service


@pytest.fixture
def mlst_reader_single_scheme() -> MLSTFeaturesReader:
    return MLSTTSeemannFeaturesReader(mlst_file=mlst_file_single_scheme)


@pytest.fixture
def mlst_reader_single_scheme2() -> MLSTFeaturesReader:
    return MLSTTSeemannFeaturesReader(mlst_file=mlst_file_single_scheme2)


@pytest.fixture
def mlst_reader_single_scheme3() -> MLSTFeaturesReader:
    return MLSTTSeemannFeaturesReader(mlst_file=mlst_file_single_scheme3)


@pytest.fixture
def mlst_data_package_single_scheme(mlst_reader_single_scheme) -> MLSTSampleDataPackage:
    return MLSTSampleDataPackage(mlst_reader_single_scheme)


@pytest.fixture
def mlst_data_package_single_scheme2(mlst_reader_single_scheme2) -> MLSTSampleDataPackage:
    return MLSTSampleDataPackage(mlst_reader_single_scheme2)


@pytest.fixture
def mlst_data_package_single_scheme3(mlst_reader_single_scheme3) -> MLSTSampleDataPackage:
    return MLSTSampleDataPackage(mlst_reader_single_scheme3)


@pytest.fixture
def mlst_reader_basic() -> MLSTFeaturesReader:
    return MLSTTSeemannFeaturesReader(mlst_file=basic_mlst_file)


@pytest.fixture
def mlst_data_package_basic(mlst_reader_basic) -> MLSTSampleDataPackage:
    return MLSTSampleDataPackage(mlst_reader_basic)


@pytest.fixture
def mlst_reader_unknown() -> MLSTFeaturesReader:
    return MLSTTSeemannFeaturesReader(mlst_file=mlst_file_unknown)


@pytest.fixture
def mlst_data_package_unknown(mlst_reader_unknown) -> MLSTSampleDataPackage:
    return MLSTSampleDataPackage(mlst_reader_unknown)


@pytest.fixture
def mlst_service_loaded(mlst_data_package_basic, database, sample_service, filesystem_storage) -> MLSTService:
    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)
    mlst_service.insert(data_package=mlst_data_package_basic)

    return mlst_service


@pytest.fixture
def mlst_service_loaded_unknown(mlst_data_package_unknown, database, sample_service, filesystem_storage) -> MLSTService:
    mlst_service = MLSTService(database_connection=database,
                               sample_service=sample_service,
                               mlst_dir=filesystem_storage.mlst_dir)
    mlst_service.insert(data_package=mlst_data_package_unknown)

    return mlst_service



import tempfile
import warnings
from pathlib import Path

import pytest
from ete3 import Tree

warnings.filterwarnings("ignore", category=DeprecationWarning)

from genomics_data_index.configuration.Project import Project
from genomics_data_index.storage.io.mlst.MLSTSampleDataPackage import MLSTSampleDataPackage
from genomics_data_index.storage.io.mlst.MLSTTSeemannFeaturesReader import MLSTTSeemannFeaturesReader
from genomics_data_index.api.query.GenomicsDataIndex import GenomicsDataIndex

from genomics_data_index.test.integration import sample_dirs, reference_file, basic_mlst_file, tree_file, \
    mlst_snippy_file
from genomics_data_index.test.integration import sourmash_signatures, snpeff_sample_vcfs, snpeff_sample_vcfs_fake_dup, \
    reference_file_snpeff, snpeff_sample_beds
from genomics_data_index.test.integration import snippy_sample2_vcfs_dict, snippy_sample2_mask_sequences_dict
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor
from genomics_data_index.storage.io.mutation.variants_processor.MultipleProcessVcfVariantsTableProcessor import \
    MultipleProcessVcfVariantsTableProcessorFactory


@pytest.fixture
def snippy_data_package_2() -> NucleotideSampleDataPackage:
    tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snippy_sample2_vcfs_dict,
                                                                     masked_genomic_files_map=snippy_sample2_mask_sequences_dict,
                                                                     sample_files_processor=SerialSampleFilesProcessor(
                                                                         tmp_dir))
    return data_package


@pytest.fixture
def loaded_database_connection() -> DataIndexConnection:
    snippy_tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                  sample_files_processor=SerialSampleFilesProcessor(
                                                                      snippy_tmp_dir))
    return build_database_connection(data_package=data_package)


@pytest.fixture
def loaded_database_connection_parallel_variants() -> DataIndexConnection:
    snippy_tmp_dir = Path(tempfile.mkdtemp())
    variants_processor_factory = MultipleProcessVcfVariantsTableProcessorFactory(ncores=2)
    data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                  sample_files_processor=SerialSampleFilesProcessor(
                                                                      snippy_tmp_dir),
                                                                  variants_processor_factory=variants_processor_factory)
    return build_database_connection(data_package=data_package)


def build_database_connection(data_package: NucleotideSampleDataPackage) -> DataIndexConnection:
    tmp_dir = Path(tempfile.mkdtemp())
    database_connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                      database_dir=tmp_dir)

    # Load Nucleotide variation
    database_connection.reference_service.add_reference_genome(reference_file)
    database_connection.variation_service.insert(data_package, feature_scope_name='genome')

    # Load MLST
    mlst_package = MLSTSampleDataPackage(MLSTTSeemannFeaturesReader(mlst_file=basic_mlst_file))
    database_connection.mlst_service.insert(mlst_package)

    # Load kmers
    kmer_service = database_connection.kmer_service
    for sample_name in sourmash_signatures:
        kmer_service.insert_kmer_index(sample_name=sample_name,
                                       kmer_index_path=sourmash_signatures[sample_name])

    # Load MLST results with overlapping samples with Nucleotide variation
    mlst_package_snippy = MLSTSampleDataPackage(MLSTTSeemannFeaturesReader(mlst_file=mlst_snippy_file))
    database_connection.mlst_service.insert(mlst_package_snippy)

    return database_connection


@pytest.fixture
def loaded_database_connection_annotations() -> DataIndexConnection:
    tmp_dir = Path(tempfile.mkdtemp())
    database_connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                      database_dir=tmp_dir)

    # Load Nucleotide variation
    database_connection.reference_service.add_reference_genome(reference_file_snpeff)
    vcf_tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snpeff_sample_vcfs,
                                                                     masked_genomic_files_map=None,
                                                                     sample_files_processor=SerialSampleFilesProcessor(
                                                                         vcf_tmp_dir))
    database_connection.variation_service.insert(data_package, feature_scope_name='NC_011083')

    return database_connection


@pytest.fixture
def loaded_database_connection_annotations_unknown() -> DataIndexConnection:
    tmp_dir = Path(tempfile.mkdtemp())
    database_connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                      database_dir=tmp_dir)

    # Load Nucleotide variation
    database_connection.reference_service.add_reference_genome(reference_file_snpeff)
    vcf_tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snpeff_sample_vcfs,
                                                                     masked_genomic_files_map=snpeff_sample_beds,
                                                                     sample_files_processor=SerialSampleFilesProcessor(
                                                                         vcf_tmp_dir))
    database_connection.variation_service.insert(data_package, feature_scope_name='NC_011083')

    return database_connection


@pytest.fixture
def loaded_database_connection_annotations_duplicate_genes() -> DataIndexConnection:
    tmp_dir = Path(tempfile.mkdtemp())
    database_connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                      database_dir=tmp_dir)

    # Load Nucleotide variation
    database_connection.reference_service.add_reference_genome(reference_file_snpeff)
    vcf_tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snpeff_sample_vcfs_fake_dup,
                                                                     masked_genomic_files_map=None,
                                                                     sample_files_processor=SerialSampleFilesProcessor(
                                                                         vcf_tmp_dir))
    database_connection.variation_service.insert(data_package, feature_scope_name='NC_011083')

    return database_connection


@pytest.fixture
def loaded_database_genomic_data_store(loaded_database_connection):
    return GenomicsDataIndex(connection=loaded_database_connection)


@pytest.fixture
def loaded_database_genomic_data_store_annotations(loaded_database_connection_annotations):
    return GenomicsDataIndex(connection=loaded_database_connection_annotations)


@pytest.fixture
def loaded_database_genomic_data_store_annotations_include_unknown(loaded_database_connection_annotations_unknown):
    return GenomicsDataIndex(connection=loaded_database_connection_annotations_unknown)


@pytest.fixture
def loaded_database_genomic_data_store_with_tree(loaded_database_connection_with_built_tree):
    return GenomicsDataIndex(connection=loaded_database_connection_with_built_tree)


@pytest.fixture
def loaded_database_connection_with_built_tree() -> DataIndexConnection:
    tmp_dir = Path(tempfile.mkdtemp())
    database_connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                      database_dir=tmp_dir)

    # Load Nucleotide variation
    database_connection.reference_service.add_reference_genome(reference_file)
    snippy_tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                  SerialSampleFilesProcessor(snippy_tmp_dir))
    database_connection.variation_service.insert(data_package, feature_scope_name='genome')
    database_connection.tree_service.rebuild_tree(reference_name='genome', align_type='full')

    # Load MLST
    mlst_package = MLSTSampleDataPackage(MLSTTSeemannFeaturesReader(mlst_file=basic_mlst_file))
    database_connection.mlst_service.insert(mlst_package)

    return database_connection


@pytest.fixture
def loaded_database_only_snippy() -> DataIndexConnection:
    tmp_dir = Path(tempfile.mkdtemp())
    database_connection = DataIndexConnection.connect(database_connection='sqlite:///:memory:',
                                                      database_dir=tmp_dir)

    # Load Nucleotide variation
    database_connection.reference_service.add_reference_genome(reference_file)
    snippy_tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                  SerialSampleFilesProcessor(snippy_tmp_dir))
    database_connection.variation_service.insert(data_package, feature_scope_name='genome')

    return database_connection


@pytest.fixture
def loaded_data_store_from_project_dir() -> GenomicsDataIndex:
    tmp_dir = Path(tempfile.mkdtemp())
    project_dir = tmp_dir / 'project'
    project = Project.initialize_project(project_dir)

    database_connection = project.create_connection()

    # Load Nucleotide variation
    database_connection.reference_service.add_reference_genome(reference_file)
    snippy_tmp_dir = Path(tempfile.mkdtemp())
    data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs,
                                                                  SerialSampleFilesProcessor(snippy_tmp_dir))
    database_connection.variation_service.insert(data_package, feature_scope_name='genome')

    # Load MLST
    mlst_package = MLSTSampleDataPackage(MLSTTSeemannFeaturesReader(mlst_file=basic_mlst_file))
    database_connection.mlst_service.insert(mlst_package)
    print(tmp_dir)

    return GenomicsDataIndex.connect(project_dir=project_dir)


@pytest.fixture
def prebuilt_tree() -> Tree:
    return Tree(str(tree_file))



import abc
from typing import List

import pandas as pd

from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet, AllSampleSet


class FeaturesComparator(abc.ABC):
    FEATURES_SELECTIONS = ['all', 'unique']

    def __init__(self, connection: DataIndexConnection, include_unknown_samples: bool,
                 include_unknown_no_present_samples: bool):
        self._connection = connection
        self._include_unknown_samples = include_unknown_samples
        self._include_unknown_no_present_samples = include_unknown_no_present_samples

    def _get_total(self, samples: SampleSet) -> int:
        if isinstance(samples, AllSampleSet):
            return self._connection.sample_service.count_samples()
        else:
            return len(samples)

    @property
    @abc.abstractmethod
    def feature_id_columns(self) -> List[str]:
        pass

    @property
    @abc.abstractmethod
    def summary_columns(self) -> List[str]:
        pass

    @property
    @abc.abstractmethod
    def index_name(self) -> str:
        pass

    def _join_additional_columns(self, features_df: pd.DataFrame) -> pd.DataFrame:
        return features_df

    @abc.abstractmethod
    def features_comparison(self, selected_samples: SampleSet,
                            sample_categories: List[SampleSet],
                            category_prefixes: List[str] = None,
                            category_samples_threshold: int = None,
                            use_only_samples_in_categories: bool = True,
                            unit: str = 'percent') -> pd.DataFrame:
        """
        Creates a dataframe which compares different categories of samples with each other with respect to features.

        For example, if kind=='mutations', compare_kind == 'percent' and there are two sample_categories then
        this will return dataframe like:

        | Mutation     | Total | Category1_percent | Category2_percent | Category1_total | Category2_total |
        | ref:100:A:T  | 10    | 50%               | 100%              | 8               | 2               |
        | ref:200:CT:C | 10    | 100%              | 0%                | 8               | 2               |
        | ...          | ...   | ...               | ...               | 8               | 2               |

        Here, "Category1_percent" is the percent of samples in Category1 that have this mutation/feature
        (50% or 4 out of 8 samples in Category1). "Category2_percent" is the percent of samples in Category2 with the
        feature (100% or 2 out of 2 samples in Category2).

        "Category1_total" and "Category2_total" are the total samples in each category. "Total" is the total
        samples in the overall query that form the universe from which we are defining "Category1" and "Category2".

        Note: since categories are defined based on sample sets, there is no enforcement that categories are
        mutually exclusive (that is, "Category1_total" + "Category2_total" will not always equal "Total"). This
        is done on purpose in case the categories you wish to compare are not mutually exclusive.

        :param selected_samples: The set of selected samples of which sample_categories will form subsets of.
        :param sample_categories: The different categories to compare.
        :param category_prefixes: The prefixes to use for the different categories (defaults to Category1, Category2, ...).
        :param category_samples_threshold: A threshold on the number of samples in a category for it to be considered.
        :param use_only_samples_in_categories: Whether or not to only use samples in categories for comparison
                                               or to use all selected samples. If all selected samples are used
                                               then it's possible to include features in the comparison table
                                               where no samples are found in any category.
        :param unit: The type of data to compare in each category (either 'percent', 'proportion', or 'count').
        :return: A dataframe comparing each category with respect to the differences in features.
        """
        pass

    @abc.abstractmethod
    def summary(self, sample_set: SampleSet) -> pd.DataFrame:
        """
        Given a samples set, summarizes the features for all samples in this set.
        :param sample_set: The set of samples to summarize features in.
        :return: A dataframe summarizing features in this set of samples.
        """
        pass

    def unique_summary(self, sample_set: SampleSet, other_set: SampleSet) -> pd.DataFrame:
        """
        Given a samples set, summarizes the features for all samples in this set that are not in another set
        (i.e., are unique to the given sample set compared to the other set).
        :param sample_set: The set of samples to summarize features in.
        :param other_set: The set of samples features should not appear in.
        :return: A dataframe summarizing unique features in this set of samples.
        """
        features_df = self.summary(sample_set)
        features_complement_df = self.summary(other_set)
        features_merged_df = features_df.merge(features_complement_df, left_index=True, right_index=True,
                                               how='left', indicator=True, suffixes=('_x', '_y'))
        rename_dict = {col + '_x': col for col in self.summary_columns}
        features_unique_df = features_merged_df[features_merged_df['_merge'] == 'left_only'].rename(rename_dict,
                                                                                                    axis='columns')
        features_unique_df = features_unique_df[self.summary_columns]
        return self._join_additional_columns(features_unique_df)



from genomics_data_index.storage.service import SQLQueryInBatcherDict, SQLQueryInBatcherList


def test_sql_query_in_batcher_dict():
    in_data = ['A', 'B', 'C', 'D', 'E']

    # Test batch size 1
    batcher = SQLQueryInBatcherDict(in_data=in_data, batch_size=1)
    results = batcher.process(lambda in_batch: {x: True for x in in_batch})
    assert isinstance(results, dict)
    assert 5 == len(results)
    assert {'A', 'B', 'C', 'D', 'E'} == set(results.keys())

    # Test batch size 2
    batcher = SQLQueryInBatcherDict(in_data=in_data, batch_size=2)
    results = batcher.process(lambda in_batch: {x: True for x in in_batch})
    assert isinstance(results, dict)
    assert 5 == len(results)
    assert {'A', 'B', 'C', 'D', 'E'} == set(results.keys())

    # Test batch size 5
    batcher = SQLQueryInBatcherDict(in_data=in_data, batch_size=5)
    results = batcher.process(lambda in_batch: {x: True for x in in_batch})
    assert isinstance(results, dict)
    assert 5 == len(results)
    assert {'A', 'B', 'C', 'D', 'E'} == set(results.keys())

    # Test batch size 6
    batcher = SQLQueryInBatcherDict(in_data=in_data, batch_size=6)
    results = batcher.process(lambda in_batch: {x: True for x in in_batch})
    assert isinstance(results, dict)
    assert 5 == len(results)
    assert {'A', 'B', 'C', 'D', 'E'} == set(results.keys())


def test_sql_query_in_batcher_list():
    in_data = ['A', 'B', 'C', 'D', 'E']

    # Test batch size 1
    batcher = SQLQueryInBatcherList(in_data=in_data, batch_size=1)
    results = batcher.process(lambda in_batch: [x for x in in_batch])
    assert isinstance(results, list)
    assert 5 == len(results)
    assert in_data == results

    # Test batch size 2
    batcher = SQLQueryInBatcherList(in_data=in_data, batch_size=2)
    results = batcher.process(lambda in_batch: in_batch)
    assert isinstance(results, list)
    assert 5 == len(results)
    assert in_data == results

    # Test batch size 5
    batcher = SQLQueryInBatcherList(in_data=in_data, batch_size=5)
    results = batcher.process(lambda in_batch: in_batch)
    assert isinstance(results, list)
    assert 5 == len(results)
    assert in_data == results

    # Test batch size 6
    batcher = SQLQueryInBatcherList(in_data=in_data, batch_size=6)
    results = batcher.process(lambda in_batch: in_batch)
    assert isinstance(results, list)
    assert 5 == len(results)
    assert in_data == results



import abc
import logging
import os
import time
from pathlib import Path
from typing import List, Union, Tuple, Dict

from genomics_data_index.storage.util import execute_commands

logger = logging.getLogger(__name__)


class KmerIndexer(abc.ABC):

    def __init__(self):
        pass

    @abc.abstractmethod
    def index(self, index_name: str, index_path: Path, files: List[Path]):
        pass

    @abc.abstractmethod
    def is_compress(self):
        pass


class KmerIndexerSourmash(KmerIndexer):

    def __init__(self, k: Union[int, List[int]] = 31, scaled: int = None, num: int = None,
                 abund: bool = False, compress: bool = True):
        super().__init__()

        if k is None:
            raise Exception('Cannot index with k=None')
        elif not isinstance(k, list) and not isinstance(k, int):
            raise Exception(f'Invalid parameter k={k}')

        if scaled is None and num is None:
            raise Exception('Must set one of "scaled" or "num" to index')
        elif scaled is not None and num is not None:
            raise Exception(f'Cannot set both scaled={scaled} and num={num}')

        self._k = k
        self._scaled = scaled
        self._num = num
        self._abund = abund

        self._params = ''
        if isinstance(k, list):
            self._params += ','.join([f'k={v}' for v in k])
        else:
            self._params += f'k={k}'

        if scaled is not None:
            self._params += f',scaled={scaled}'
        else:
            self._params += f',num={num}'

        if abund:
            self._params += ',abund'

        self._compress = compress

    def is_compress(self):
        return self._compress

    def index(self, index_name: str, index_path: Path, files: List[Path]) -> Path:
        if self._compress:
            if index_path.suffix != '.gz':
                logger.warning((f'index_path=[{index_path}] does not end in ".gz" but compress={self._compress}. '
                                'Adding ".gz" to index_path.'))
                index_path = Path(str(index_path) + '.gz')
            index_out = Path(str(index_path) + '.tmp')
        else:
            index_out = index_path

        command = ['sourmash', 'sketch', 'dna', '-p', self._params,
                   '--name', index_name, '--output', str(index_out)]
        command.extend([str(f) for f in files])
        if self._compress:
            execute_commands([command, ['gzip', str(index_out)]])
            index_gzip_out = Path(str(index_out) + '.gz')
            os.rename(index_gzip_out, index_path)
        else:
            execute_commands([command])

        return index_path


class KmerIndexManager:

    def __init__(self, kmer_index_directory: Path, kmer_indexer: KmerIndexer):
        self._kmer_index_directory = kmer_index_directory
        self._kmer_indexer = kmer_indexer

    def index_single_genome(self, index_name: str, files: List[Path]) -> Path:
        """
        Indexes reads and constructs an index file with the given name from the given files.
        :param index_name: The name of the index.
        :param files: The list of files to index.
        :return: The path to the index.
        """
        if self._kmer_indexer.is_compress():
            index_out = self._kmer_index_directory / f'{index_name}.sig.gz'
        else:
            index_out = self._kmer_index_directory / f'{index_name}.sig'

        if index_out.exists():
            raise Exception(f'Index output file {index_out} already exists')

        indexed_path = self._kmer_indexer.index(index_name=index_name,
                                                index_path=index_out,
                                                files=files)

        return indexed_path

    def index_all_genomes(self, genomes_files: List[Tuple[str, List[Path]]]) -> Dict[str, Path]:
        start_time = time.time()
        logger.debug('Start building kmer indexes')

        indexed_genomes = {}

        for genome_files in genomes_files:
            index_name = genome_files[0]
            index_files = genome_files[1]
            logger.debug(f'Working on {index_name}')
            indexed_path = self.index_single_genome(index_name=index_name, files=index_files)
            logger.debug(f'Finished creating index {indexed_path}')

            indexed_genomes[index_name] = indexed_path

        end_time = time.time()
        logger.debug(f'Finished building kmer indexes. Took {end_time - start_time:0.2f} seconds')

        return indexed_genomes



from pathlib import Path

import pandas as pd

from genomics_data_index.storage.io.SampleData import SampleData


class MLSTSampleData(SampleData):

    def __init__(self, sample_name: str, scheme: str, mlst_allele_data: pd.Series):
        super().__init__(sample_name=sample_name)
        self._mlst_allele_data = mlst_allele_data
        self._scheme = scheme

    def get_scheme(self) -> str:
        return self._scheme

    def get_alleles(self):
        return self._mlst_allele_data

    def is_preprocessed(self) -> bool:
        return True

    def _do_preprocess_and_persist(self, output_dir: Path) -> SampleData:
        return self

    def _do_persist(self, output_dir: Path) -> SampleData:
        return self
